/* Scrutari - Copyright (c) 2009-2017 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.sctserver.servlets.webapp.htmlpages;

import fr.exemole.sctserver.api.log.SctLogConstants;
import fr.exemole.sctserver.servlets.ScrutariWebapp;
import fr.exemole.sctserver.tools.EngineUtils;
import java.io.IOException;
import java.util.List;
import javax.servlet.http.HttpServletResponse;
import net.mapeadores.util.buildinfo.BuildInfo;
import net.mapeadores.util.html.HA;
import net.mapeadores.util.html.HtmlAttributes;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.localisation.LocalisationUtils;
import net.mapeadores.util.logging.CategorizedMessageEntry;
import net.mapeadores.util.logging.MessageLog;
import net.mapeadores.util.primitives.FuzzyDate;
import net.mapeadores.util.text.DateFormatBundle;
import net.mapeadores.util.text.Label;
import net.mapeadores.util.text.StringUtils;


/**
 *
 * @author Vincent Calame
 */
public class IndexPage extends WebappHtmlPage {

    private final static HtmlAttributes WEBAPP_ERROR_ATTR = HA.classes("scrutari-webapp-Error");

    private final boolean afterReload;

    public IndexPage(ScrutariWebapp scrutariWebapp, Lang workingLang, boolean afterReload) {
        super(scrutariWebapp, workingLang);
        this.afterReload = afterReload;
    }

    @Override
    public void handleResponse(HttpServletResponse response) throws IOException {
        initResponse(response);
        start();
        startBSContainer();
        printLogo();
        if (!scrutariWebapp.isInit()) {
            printInitErrors();
        } else {
            printInfo();
            printConfLog();
            printEngineList();
        }
        printBuildInfo();
        endBSContainer();
        end();
    }

    private void printLogo() {
        startBSRow();
        this
                .IMG(HA.src("_resources/theme/images/logo.png").classes("scrutari-webapp-Logo"));
        endBSRow();
    }

    private void printInfo() {
        startBSRow();
        String locKey = (afterReload) ? "_ info.webapp.updatedone" : "_ info.webapp.congratulations";
        this
                .P(HA.classes("scrutari-webapp-Title"))
                .localize(locKey)
                ._P();
        endBSRow();
    }

    private void printInitErrors() {
        startBSRow();
        List<MessageLog.LogGroup> logGroupList = scrutariWebapp.getInitMessageLog().getLogGroupList();
        this
                .P(HA.classes("scrutari-webapp-ErrorTitle"))
                .localize("_ info.webapp.initerrors")
                ._P();
        this
                .UL();
        for (MessageLog.LogGroup logGroup : logGroupList) {
            this
                    .LI();
            this
                    .P()
                    .CODE()
                    .escape('[')
                    .escape(logGroup.getURI())
                    .escape(']')
                    ._CODE()
                    ._P();
            this
                    .UL();
            for (CategorizedMessageEntry entry : logGroup.getEntryList()) {
                this
                        .LI()
                        .P()
                        .localize(entry.getMessage())
                        ._P()
                        ._LI();
            }
            this
                    ._UL();
            this
                    ._LI();
        }
        this
                ._UL();
        endBSRow();
    }

    private void printConfLog() {
        List<MessageLog.LogGroup> logGroupList = scrutariWebapp.getConfMessageLog().getLogGroupList();
        if (logGroupList.isEmpty()) {
            return;
        }
        startBSRow();
        this
                .P(HA.classes("scrutari-webapp-ErrorTitle"))
                .localize("_ info.webapp.conferrors")
                ._P();
        this
                .UL();
        for (MessageLog.LogGroup logGroup : logGroupList) {
            this
                    .LI();
            this
                    .P()
                    .CODE()
                    .escape('[')
                    .escape(logGroup.getURI())
                    .escape(']')
                    ._CODE()
                    ._P();
            this
                    .UL();
            for (CategorizedMessageEntry entry : logGroup.getEntryList()) {
                String category = entry.getCategory();
                this
                        .LI();
                if (category.equals(SctLogConstants.CRITICAL_CATEGORY)) {
                    this
                            .P(WEBAPP_ERROR_ATTR);
                } else {
                    this
                            .P();
                }
                if (category.length() > 0) {
                    this
                            .CODE()
                            .escape('[')
                            .escape(category)
                            .escape(']')
                            ._CODE()
                            .space();
                }
                this
                        .localize(entry.getMessage())
                        ._P()
                        ._LI();
            }
            this
                    ._UL();
            this
                    ._LI();
        }
        this
                ._UL();
        endBSRow();

    }

    private void printBuildInfo() {
        BuildInfo buildInfo = scrutariWebapp.getBuildInfo();
        if (buildInfo == null) {
            return;
        }
        startBSRow();
        this
                .H3()
                .localize("_ title.webapp.buildinfo")
                ._H3();
        this
                .P()
                .localize("_ label.webapp.version")
                .colon()
                .space()
                .escape(buildInfo.getVersion())
                ._P();
        this
                .P()
                .localize("_ label.webapp.repository")
                .colon()
                .space()
                .escape(buildInfo.getRepository())
                ._P();
        String dateString = "";
        FuzzyDate date = buildInfo.getDate();
        if (date != null) {
            DateFormatBundle dateFormatBundle = DateFormatBundle.getDateFormatBundle(workingLang.toLocale());
            dateString = date.getDateLitteral(dateFormatBundle);
        }
        this
                .P()
                .localize("_ label.webapp.builddate")
                .colon()
                .space()
                .escape(dateString)
                ._P();
    }

    private void printEngineList() {
        startBSRow();
        List<String> nameList = scrutariWebapp.getEngineNameList();
        int size = nameList.size();
        this
                .H3();
        if (size == 0) {
            this
                    .localize("_ info.webapp.engine_none");
        } else if (size == 1) {
            this
                    .localize("_ info.webapp.engine_one");
        } else {
            this
                    .localize(LocalisationUtils.toMessage("_ info.webapp.engine_many", StringUtils.toString(size, workingLang.toLocale())));
        }
        this
                ._H3();
        if (size > 0) {
            this
                    .UL();
            for (String name : nameList) {
                String title = "";
                if (scrutariWebapp.isEngineInit(name)) {
                    Label titleLabel = EngineUtils.getTitleLabel(scrutariWebapp.getEngine(name), workingLang);
                    if (titleLabel != null) {
                        title = " – " + titleLabel.getLabelString();
                    }
                }
                this
                        .LI()
                        .P()
                        .A(HA.href(name + "/admin/"))
                        .CODE()
                        .escape("[")
                        .escape(name)
                        .escape("]")
                        ._CODE()
                        .escape(title)
                        ._A()
                        ._P()
                        ._LI();
            }
            this
                    ._UL();
        }
        endBSRow();
        startBSRow();
        this
                .P()
                .A(HA.href("_reload"))
                .localize("_ link.global.reload")
                ._A()
                ._P();
        endBSRow();
    }

}
