/* ScrutariLib_Data - Copyright (c) 2005-2017 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package net.scrutari.data;

import java.io.IOException;
import java.io.Serializable;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.primitives.io.PrimitivesReader;
import net.mapeadores.util.primitives.io.PrimitivesWriter;
import net.mapeadores.util.text.Label;
import net.scrutari.datauri.CorpusURI;


/**
 *
 * @author Vincent Calame
 */
public final class CorpusData implements Serializable {

    public final static int INTITULECORPUS_FIELDNUMBER = -11;
    public final static int INTITULEFICHE_FIELDNUMBER = -22;
    private final CorpusURI corpusURI;
    private final int codecorpus;
    private final int codebase;
    private final int complementCount;
    private final CorpusLangData[] corpusLangDataArray;
    private final int[] codesFiches;

    private CorpusData(CorpusURI corpusURI, int codecorpus, int codebase, int complementCount, CorpusLangData[] corpusLangDataArray, int[] codesFiches) {
        this.corpusURI = corpusURI;
        this.codecorpus = codecorpus;
        this.codebase = codebase;
        this.complementCount = complementCount;
        this.corpusLangDataArray = corpusLangDataArray;
        this.codesFiches = codesFiches;
    }

    public CorpusURI getCorpusURI() {
        return corpusURI;
    }

    /**
     * Retourne le nombre maximum de champs complémentaires.
     */
    public int getComplementMaxNumber() {
        return complementCount;
    }

    /**
     * Retourne le code du corpus.
     */
    public int getCodeCorpus() {
        return codecorpus;
    }

    /**
     * Retourne le code de la base.
     */
    public int getCodeBase() {
        return codebase;
    }

    /**
     * Retourne le total de fiches du corpus.
     */
    public int getFicheCount() {
        return codesFiches.length;
    }

    /**
     * Retourne le code de la i-ième fiche du corpus.
     */
    public int getCodeFiche(int i) {
        return codesFiches[i];
    }

    public CorpusLangData getLangDataByLang(Lang lang) {
        int length = corpusLangDataArray.length;
        for (int i = 0; i < length; i++) {
            CorpusLangData corpusLangData = corpusLangDataArray[i];
            if (corpusLangData.lang.equals(lang)) {
                return corpusLangData;
            }
        }
        return null;
    }

    public int getLangDataCount() {
        return corpusLangDataArray.length;
    }

    public CorpusLangData getLangDataByIndex(int index) {
        return corpusLangDataArray[index];
    }

    public static CorpusData fromPrimitives(PrimitivesReader primitivesReader) throws IOException {
        CorpusURI corpusURI = (CorpusURI) URIPrimitivesUtils.readScrutariDataURI(primitivesReader);
        int codecorpus = primitivesReader.readInt();
        int codebase = primitivesReader.readInt();
        int complementCount = primitivesReader.readInt();
        int langDataLength = primitivesReader.readInt();
        CorpusLangData[] corpusLangDataArray = new CorpusLangData[langDataLength];
        for (int i = 0; i < langDataLength; i++) {
            Lang lang = Lang.build(primitivesReader.readString());
            CorpusLangData corpusLangData = new CorpusLangData(lang, complementCount);
            readLabel(corpusLangData, primitivesReader);
            readLabel(corpusLangData, primitivesReader);
            for (int j = 0; j < complementCount; j++) {
                readLabel(corpusLangData, primitivesReader);
            }
            corpusLangDataArray[i] = corpusLangData;
        }
        int fichesLength = primitivesReader.readInt();
        int[] codesFiches = new int[fichesLength];
        for (int i = 0; i < fichesLength; i++) {
            codesFiches[i] = primitivesReader.readInt();
        }
        return new CorpusData(corpusURI, codecorpus, codebase, complementCount, corpusLangDataArray, codesFiches);
    }

    public static void toPrimitives(CorpusData corpusData, PrimitivesWriter primitivesWriter) throws IOException {
        URIPrimitivesUtils.writeScrutariDataURI(corpusData.corpusURI, primitivesWriter);
        primitivesWriter.writeInt(corpusData.codecorpus);
        primitivesWriter.writeInt(corpusData.codebase);
        int complementCount = corpusData.complementCount;
        primitivesWriter.writeInt(complementCount);
        int langDataLength = corpusData.corpusLangDataArray.length;
        primitivesWriter.writeInt(langDataLength);
        for (int i = 0; i < langDataLength; i++) {
            CorpusLangData corpusLangData = corpusData.corpusLangDataArray[i];
            primitivesWriter.writeString(corpusLangData.lang.toString());
            writeLabel(INTITULECORPUS_FIELDNUMBER, corpusLangData, primitivesWriter);
            writeLabel(INTITULEFICHE_FIELDNUMBER, corpusLangData, primitivesWriter);
            for (int j = 0; j < complementCount; j++) {
                writeLabel(j + 1, corpusLangData, primitivesWriter);
            }
        }
        int fichesLength = corpusData.codesFiches.length;
        primitivesWriter.writeInt(fichesLength);
        for (int i = 0; i < fichesLength; i++) {
            primitivesWriter.writeInt(corpusData.codesFiches[i]);
        }
    }

    private static void writeLabel(int fieldNumber, CorpusLangData corpusLangData, PrimitivesWriter primitivesWriter) throws IOException {
        primitivesWriter.writeInt(fieldNumber);
        String labelString = corpusLangData.getLabelString(fieldNumber);
        primitivesWriter.writeString(labelString);
    }

    private static void readLabel(CorpusLangData corpusLangData, PrimitivesReader primitivesReader) throws IOException {
        int fieldNumber = primitivesReader.readInt();
        String labelString = primitivesReader.readString();
        if (labelString != null) {
            corpusLangData.setLabel(DataLabelBuilder.toLabel(corpusLangData.lang, labelString), fieldNumber);
        }
    }


    public static class CorpusLangData {

        private final Lang lang;
        private Label corpusLabel;
        private Label ficheLabel;
        private final Label[] complementLabelArray;

        private CorpusLangData(Lang lang, int complementCount) {
            this.lang = lang;
            complementLabelArray = new Label[complementCount];
        }

        public Label getLabel(int fieldNumber) {
            if (fieldNumber == INTITULECORPUS_FIELDNUMBER) {
                return corpusLabel;
            }
            if (fieldNumber == INTITULEFICHE_FIELDNUMBER) {
                return ficheLabel;
            }
            return complementLabelArray[fieldNumber - 1];
        }

        public String getLabelString(int fieldNumber) {
            Label label = getLabel(fieldNumber);
            if (label == null) {
                return null;
            }
            return label.getLabelString();
        }

        public Lang getLang() {
            return lang;
        }

        private void setLabel(Label label, int fieldNumber) {
            if (fieldNumber == INTITULECORPUS_FIELDNUMBER) {
                corpusLabel = label;
            } else if (fieldNumber == INTITULEFICHE_FIELDNUMBER) {
                ficheLabel = label;
            } else {
                complementLabelArray[fieldNumber - 1] = label;
            }
        }

    }


    public static class Builder {

        private final CorpusURI corpusURI;
        private final Map<Lang, CorpusLangDataBuilder> corpusLangDataMap = new LinkedHashMap<Lang, CorpusLangDataBuilder>();
        private final List<Integer> codeficheList = new ArrayList<Integer>();
        private String hrefParent;
        private int complementCount = 0;

        public Builder(CorpusURI corpusURI) {
            this.corpusURI = corpusURI;
        }

        public CorpusURI getCorpusURI() {
            return corpusURI;
        }

        public void addLabel(int fieldNumber, Label label) {
            label = DataLabelBuilder.checkLabel(label);
            if (label == null) {
                return;
            }
            if (fieldNumber > complementCount) {
                throw new IllegalArgumentException("fieldNumber > complementCount");
            }
            CorpusLangDataBuilder builder = getCorpusLangDataBuilder(label.getLang());
            builder.putLabel(fieldNumber, label);
        }

        public int createNewComplement() {
            complementCount++;
            return complementCount;
        }

        public int getComplementMaxNumber() {
            return complementCount;
        }

        public void addCodeFiche(int codefiche) {
            codeficheList.add(codefiche);
        }

        public CorpusData toCorpusData(int codecorpus, int codebase) {
            int ficheCount = codeficheList.size();
            int[] codesFiches = new int[ficheCount];
            for (int i = 0; i < ficheCount; i++) {
                codesFiches[i] = codeficheList.get(i);
            }
            int langCount = corpusLangDataMap.size();
            CorpusLangData[] corpusLangDataArray = new CorpusLangData[langCount];
            int p = 0;
            for (CorpusLangDataBuilder builder : corpusLangDataMap.values()) {
                corpusLangDataArray[p] = builder.toCorpusLangData(complementCount);
                p++;
            }
            return new CorpusData(corpusURI, codecorpus, codebase, complementCount, corpusLangDataArray, codesFiches);
        }

        public String getHrefParent() {
            return hrefParent;
        }

        public boolean setHrefParent(URI hrefParentURI) {
            if (hrefParentURI == null) {
                this.hrefParent = null;
                return true;
            }
            if (!hrefParentURI.isAbsolute()) {
                return false;
            }
            if (hrefParentURI.isOpaque()) {
                return false;
            }
            String scheme = hrefParentURI.getScheme();
            if ((!scheme.equals("http")) && (!scheme.equals("https"))) {
                return false;
            }
            this.hrefParent = hrefParentURI.toASCIIString();
            return true;
        }

        private CorpusLangDataBuilder getCorpusLangDataBuilder(Lang lang) {
            CorpusLangDataBuilder builder = corpusLangDataMap.get(lang);
            if (builder == null) {
                builder = new CorpusLangDataBuilder(lang);
                corpusLangDataMap.put(lang, builder);
            }
            return builder;
        }

    }


    private static class CorpusLangDataBuilder {

        private final Lang lang;
        private final Map<Integer, Label> labelMap = new HashMap<Integer, Label>();

        private CorpusLangDataBuilder(Lang lang) {
            this.lang = lang;
        }

        private void putLabel(int fieldNumber, Label label) {
            labelMap.put(fieldNumber, label);
        }

        private CorpusLangData toCorpusLangData(int complementCount) {
            CorpusLangData corpusLangData = new CorpusLangData(lang, complementCount);
            for (Map.Entry<Integer, Label> entry : labelMap.entrySet()) {
                corpusLangData.setLabel(entry.getValue(), entry.getKey());
            }
            return corpusLangData;
        }

    }

}
