/* ScrutariLib_SearchEngine - Copyright (c) 2008-2017 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package net.scrutari.searchengine.tools.lexieseek;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.localisation.LocalisationLexieFilters;
import net.mapeadores.util.logicaloperation.Operand;
import net.mapeadores.util.logicaloperation.SubOperand;
import net.mapeadores.util.text.collation.map.CollatedKeyMap;
import net.mapeadores.util.text.collation.map.SearchResultUnit;
import net.mapeadores.util.text.collation.map.ValueFilter;
import net.mapeadores.util.text.search.MultiSearchToken;
import net.mapeadores.util.text.search.SearchToken;
import net.mapeadores.util.text.search.SimpleSearchToken;
import net.scrutari.db.api.LexieAccess;
import net.scrutari.searchengine.api.lexieseek.LexieSeekHandler;
import net.scrutari.searchengine.api.operands.search.AlineaFilter;
import net.scrutari.searchengine.api.operands.search.SearchOperation;
import net.scrutari.searchengine.api.operands.search.SearchOperationReduction;
import net.scrutari.searchengine.api.operands.search.SearchTokenOperand;
import net.scrutari.searchengine.api.options.ComputedOptions;
import net.scrutari.searchengine.tools.operands.search.SearchOperationReductionBuilder;


/**
 *
 * @author Vincent Calame
 */
public class LexieSeeker {

    private final LexieAccess lexieAccess;
    private final Map<Integer, Integer> ficheMap = new LinkedHashMap<Integer, Integer>();
    private final Map<Integer, Integer> motcleMap = new LinkedHashMap<Integer, Integer>();
    private final ComputedOptions computedOptions;
    private final LexieSeekHandler lexieSeekHandler;
    private final ValueFilter valueFilter = null;

    public LexieSeeker(ComputedOptions computedOptions, LexieSeekHandler lexieSeekHandler, LexieAccess lexieAccess) {
        this.lexieAccess = lexieAccess;
        this.computedOptions = computedOptions;
        this.lexieSeekHandler = lexieSeekHandler;
    }

    public void seek(SearchOperation searchOperation) {
        Lang[] langArray = computedOptions.getLangArray();
        for (Lang lang : langArray) {
            SearchOperationReduction operationReduction = SearchOperationReductionBuilder.lexieReduction(searchOperation, LocalisationLexieFilters.getLexieFilter(lang));
            ficheMap.clear();
            motcleMap.clear();
            Operand operand = operationReduction.getRootOperand();
            if (operand == null) {
                continue;
            }
            seekOperand(operand, lang);
            int unusedCount = operationReduction.getUnusedNumberCount();
            if (unusedCount > 0) {
                int[] unusedArray = new int[unusedCount];
                for (int i = 0; i < unusedCount; i++) {
                    unusedArray[i] = operationReduction.getUnusedNumber(i);
                }
                for (Map.Entry<Integer, Integer> entry : ficheMap.entrySet()) {
                    int codefiche = entry.getKey();
                    int codecorpus = entry.getValue();
                    for (int j = 0; j < unusedCount; j++) {
                        lexieSeekHandler.addAlineaOccurrence(codefiche, codecorpus, unusedArray[j], -1, null);
                    }
                }
                for (Map.Entry<Integer, Integer> entry : motcleMap.entrySet()) {
                    int codemotcle = entry.getKey();
                    int codethesaurus = entry.getValue();
                    for (int j = 0; j < unusedCount; j++) {
                        lexieSeekHandler.addLangOccurrence(codemotcle, codethesaurus, unusedArray[j], lang, null, 0.0f);
                    }
                }
            }
        }
    }

    private void seekOperand(Operand operand, Lang lang) {
        if (operand instanceof SubOperand) {
            seekSubOperand((SubOperand) operand, lang);
        } else {
            SearchTokenOperand searchOperand = (SearchTokenOperand) operand;
            AlineaFilter alineaFilter = searchOperand.getAlineaFilter();
            int operandNumber = searchOperand.getOperandNumber();
            SearchToken token = searchOperand.getSearchToken();
            if (token instanceof MultiSearchToken) {
                MultiSearchToken multiToken = (MultiSearchToken) token;
                int count = multiToken.getSubtokenCount();
                if (count > 1) {
                    MultiTokenSeeker.seek(multiToken, operandNumber, lang, this, alineaFilter);
                } else if (count == 1) {
                    SimpleTokenSeeker.seek(multiToken.getSubtoken(0), operandNumber, lang, this, alineaFilter);
                }
            } else {
                SimpleTokenSeeker.seek((SimpleSearchToken) token, operandNumber, lang, this, alineaFilter);
            }
        }
    }

    private void seekSubOperand(SubOperand subOperand, Lang lang) {
        int count = subOperand.getOperandCount();
        for (int i = 0; i < count; i++) {
            seekOperand(subOperand.getOperand(i), lang);
        }
    }

    LexieAccess getLexieAccess() {
        return lexieAccess;
    }

    ComputedOptions getComputedOptions() {
        return computedOptions;
    }

    LexieSeekHandler getLexieSeekHandler() {
        return lexieSeekHandler;
    }

    void addFiche(int codefiche, int codecorpus) {
        ficheMap.put(codefiche, codecorpus);
    }

    void addMotcle(int codemotcle, int codethesaurus) {
        motcleMap.put(codemotcle, codethesaurus);
    }

    List<SearchResultUnit<Long>> search(String searchText, Lang lang, int type) {
        CollatedKeyMap<Long> collatedStringMap = lexieAccess.getLexieIdMapByLang(lang);
        if (collatedStringMap == null) {
            return null;
        }
        return collatedStringMap.search(searchText, type, valueFilter);
    }

}
