/* ScrutariLib_SearchEngine - Copyright (c) 2005-2017 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package net.scrutari.searchengine.tools.options;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import net.mapeadores.util.logging.MessageHandler;
import net.scrutari.db.api.DataAccess;
import net.scrutari.db.api.URITreeProvider;
import net.scrutari.db.api.codes.Codes;
import net.scrutari.db.api.codes.ElementCodes;
import net.scrutari.db.tools.codes.CodesUtils;
import net.scrutari.searchengine.api.operands.eligibility.Eligibility;
import net.scrutari.searchengine.api.operands.eligibility.EligibilityOperand;
import net.scrutari.searchengine.api.options.SearchOptions;
import net.scrutari.searchengine.api.options.SearchOptionsDef;
import net.scrutari.searchengine.api.result.FicheSearchResultHolder;
import net.scrutari.searchengine.tools.operands.eligibility.EligibilityCheck;
import net.scrutari.searchengine.tools.operands.eligibility.EligibilityCheckFactory;
import net.scrutari.searchengine.tools.operands.eligibility.EligibilityCheckParameters;


/**
 *
 *
 * @author Vincent Calame
 */
public final class SearchOptionsBuilder {

    public final static SearchOptions EMPTY_SEARCHOPTIONS = new InternalSearchOptions(SearchOptionsDefBuilder.EMPTY_SEARCHOPTIONSDEF, null, null, null, null);
    private final SearchOptionsDef searchOptionsDef;
    private Eligibility ficheEligibility;
    private Set<Integer> corpusSet = null;
    private Set<Integer> thesaurusSet = null;
    private Map<Integer, Integer> ficheMap = null;

    public SearchOptionsBuilder(SearchOptionsDef searchOptionsDef) {
        this.searchOptionsDef = searchOptionsDef;
    }

    public SearchOptionsDef getSearchOptionsDef() {
        return searchOptionsDef;
    }

    public void reduceFicheCodes(ElementCodes ficheCodes) {
        if (ficheCodes == null) {
            return;
        }
        if ((ficheMap != null) && (ficheMap.isEmpty())) {
            return;
        }
        if ((corpusSet != null) && (corpusSet.isEmpty())) {
            ficheMap = Collections.emptyMap();
            return;
        }
        int ficheCount = ficheCodes.getCodeCount();
        if (ficheCount == 0) {
            corpusSet = Collections.emptySet();
            ficheMap = Collections.emptyMap();
            return;
        }
        if (corpusSet == null) {
            corpusSet = new LinkedHashSet<Integer>();
            ficheMap = new LinkedHashMap<Integer, Integer>();
            for (int i = 0; i < ficheCount; i++) {
                int codecorpus = ficheCodes.getSubsetCode(i);
                corpusSet.add(codecorpus);
                ficheMap.put(ficheCodes.getCode(i), codecorpus);
            }
            return;
        }
        if (ficheMap == null) {
            ficheMap = new LinkedHashMap<Integer, Integer>();
            for (int i = 0; i < ficheCount; i++) {
                int codecorpus = ficheCodes.getSubsetCode(i);
                if (corpusSet.contains(codecorpus)) {
                    ficheMap.put(ficheCodes.getCode(i), codecorpus);
                }
            }
            return;
        }
        Set<Integer> newCorpusSet = new HashSet<Integer>();
        Map<Integer, Integer> newFicheMap = new LinkedHashMap<Integer, Integer>();
        for (int i = 0; i < ficheCount; i++) {
            int codecorpus = ficheCodes.getSubsetCode(i);
            if (corpusSet.contains(codecorpus)) {
                int codefiche = ficheCodes.getCode(i);
                if (ficheMap.containsKey(codefiche)) {
                    newCorpusSet.add(codecorpus);
                    newFicheMap.put(codefiche, codecorpus);
                }
            }
        }
        corpusSet = newCorpusSet;
        ficheMap = newFicheMap;
    }

    public void reduceCorpusCodes(Codes corpusCodes) {
        if (corpusCodes == null) {
            return;
        }
        if ((corpusSet != null) && (corpusSet.isEmpty())) {
            return;
        }
        int corpusCount = corpusCodes.getCodeCount();
        if (corpusCount == 0) {
            corpusSet = Collections.emptySet();
            if (ficheMap != null) {
                ficheMap = Collections.emptyMap();
            }
            return;
        }
        if (corpusSet == null) {
            corpusSet = new LinkedHashSet<Integer>();
            int count = corpusCodes.getCodeCount();
            for (int i = 0; i < count; i++) {
                corpusSet.add(corpusCodes.getCode(i));
            }
            return;
        }
        Set<Integer> newCorpusSet = new HashSet<Integer>();
        for (int i = 0; i < corpusCount; i++) {
            int code = corpusCodes.getCode(i);
            if (corpusSet.contains(code)) {
                newCorpusSet.add(code);
            }
        }
        if (ficheMap != null) {
            for (Iterator<Map.Entry<Integer, Integer>> it = ficheMap.entrySet().iterator(); it.hasNext();) {
                Map.Entry<Integer, Integer> entry = it.next();
                if (!newCorpusSet.contains(entry.getValue())) {
                    it.remove();
                }
            }
        }
        corpusSet = newCorpusSet;
    }

    public void reduceThesaurusCodes(Codes thesaurusCodes) {
        if (thesaurusCodes == null) {
            return;
        }
        if ((thesaurusSet != null) && (thesaurusSet.isEmpty())) {
            return;
        }
        int count = thesaurusCodes.getCodeCount();
        if (count == 0) {
            thesaurusSet = Collections.emptySet();
            return;
        }
        if (thesaurusSet == null) {
            thesaurusSet = new LinkedHashSet<Integer>();
            for (int i = 0; i < count; i++) {
                thesaurusSet.add(thesaurusCodes.getCode(i));
            }
            return;
        }
        Set<Integer> newThesaurusSet = new HashSet<Integer>();
        for (int i = 0; i < count; i++) {
            int code = thesaurusCodes.getCode(i);
            if (thesaurusSet.contains(code)) {
                newThesaurusSet.add(code);
            }
        }
        thesaurusSet = newThesaurusSet;
    }

    public Codes getCurrentCorpusCodes() {
        if (corpusSet != null) {
            return CodesUtils.toCodes(corpusSet);
        } else {
            return null;
        }
    }

    public SearchOptions toSearchOptions() {
        ElementCodes ficheCodes = null;
        Codes corpusCodes = null;
        Codes thesaurusCodes = null;
        if (ficheMap != null) {
            ficheCodes = CodesUtils.toElementCodes(ficheMap);
        }
        if (corpusSet != null) {
            corpusCodes = CodesUtils.toCodes(corpusSet);
        }
        if (thesaurusSet != null) {
            thesaurusCodes = CodesUtils.toCodes(thesaurusSet);
        }
        return new InternalSearchOptions(searchOptionsDef, ficheEligibility, ficheCodes, corpusCodes, thesaurusCodes);
    }

    public void checkEligibility(DataAccess dataAccess, FicheSearchResultHolder ficheSearchResultHolder, MessageHandler messageHandler, URITreeProvider uriTreeProvider) {
        EligibilityCheckParameters eligibilityCheckParameters = new EligibilityCheckParameters()
                .setDataAccess(dataAccess)
                .setFicheSearchResultHolder(ficheSearchResultHolder)
                .setMessageHandler(messageHandler)
                .setUriTreeProvider(uriTreeProvider);
        EligibilityOperand ficheEligibilityOperand = searchOptionsDef.getFicheEligibilityOperand();
        if (ficheEligibilityOperand == null) {
            return;
        }
        EligibilityCheck eligibilityCheck = EligibilityCheckFactory.check(ficheEligibilityOperand, getCurrentCorpusCodes(), eligibilityCheckParameters);
        if (eligibilityCheck == null) {
            return;
        }
        Set<Integer> codeSet = eligibilityCheck.getCodeSet();
        if (codeSet != null) {
            if (corpusSet == null) {
                corpusSet = new HashSet<Integer>(codeSet);
            } else {
                corpusSet.retainAll(codeSet);
            }
        }
        this.ficheEligibility = eligibilityCheck.getEligibility();
    }

    public static SearchOptionsBuilder derive(SearchOptions searchOptions) {
        SearchOptionsBuilder builder = new SearchOptionsBuilder(searchOptions.getSearchOptionsDef());
        builder.reduceFicheCodes(searchOptions.getFicheCodes());
        builder.reduceCorpusCodes(searchOptions.getCorpusCodes());
        builder.reduceThesaurusCodes(searchOptions.getThesaurusCodes());
        builder.ficheEligibility = searchOptions.getFicheEligibility();
        return builder;
    }


    private static class InternalSearchOptions implements SearchOptions {

        private final SearchOptionsDef searchOptionsDef;
        private final Eligibility ficheEligibility;
        private final ElementCodes ficheCodes;
        private final Codes corpusCodes;
        private final Codes thesaurusCodes;


        private InternalSearchOptions(SearchOptionsDef searchOptionsDef, Eligibility ficheEligibility, ElementCodes ficheCodes, Codes corpusCodes, Codes thesaurusCodes) {
            this.searchOptionsDef = searchOptionsDef;
            this.ficheEligibility = ficheEligibility;
            this.ficheCodes = ficheCodes;
            this.corpusCodes = corpusCodes;
            this.thesaurusCodes = thesaurusCodes;
        }

        @Override
        public SearchOptionsDef getSearchOptionsDef() {
            return searchOptionsDef;
        }

        @Override
        public Eligibility getFicheEligibility() {
            return ficheEligibility;
        }

        @Override
        public ElementCodes getFicheCodes() {
            return ficheCodes;
        }

        @Override
        public Codes getCorpusCodes() {
            return corpusCodes;
        }

        @Override
        public Codes getThesaurusCodes() {
            return thesaurusCodes;
        }

    }

}
