/* ScrutariLib_SearchEngine - Copyright (c) 2013-2017 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package net.scrutari.searchengine.tools.pertinence;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.primitives.FuzzyDate;
import net.mapeadores.util.text.SubstringPosition;
import net.scrutari.db.api.DataAccess;
import net.scrutari.lexie.AlineaInteger;
import net.scrutari.searchengine.api.pertinence.FichePertinence;
import net.scrutari.searchengine.api.pertinence.MotclePertinence;
import net.scrutari.searchengine.api.result.FicheSearchResultInfo;
import net.scrutari.searchengine.api.result.InAlineaOccurrences;
import net.scrutari.searchengine.api.result.InLangOccurrences;
import net.scrutari.searchengine.tools.result.FicheSearchResultInfoBuilder;
import net.scrutari.searchengine.tools.result.OccurrencesBuilder;


/**
 *
 * @author Vincent Calame
 */
public class PertinenceUtils {

    private PertinenceUtils() {
    }

    public static FicheSearchResultInfo toFicheSearchResultInfo(PertinencePack pertinencePack, int[] acceptOperandArray) {
        FichePertinence fichePertinence = pertinencePack.getFichePertinence();
        return FicheSearchResultInfoBuilder.toFicheSearchResultInfo(fichePertinence.getCodefiche(),
                fichePertinence.getAcceptCodemotcleArray(), getInAlineaOccurrencesArray(fichePertinence, acceptOperandArray), pertinencePack.getScoreArray());
    }

    public static InAlineaOccurrences[] getInAlineaOccurrencesArray(FichePertinence fichePertinence, int[] acceptOperandArray) {
        int acceptLength = acceptOperandArray.length;
        Map<Integer, OccurrencesBuilder> builderByAlineaMap = new LinkedHashMap<Integer, OccurrencesBuilder>();
        for (int i = 0; i < acceptLength; i++) {
            int operandNumber = acceptOperandArray[i];
            int alineaInteger = fichePertinence.getAlineaInteger(operandNumber);
            if (alineaInteger < 0) {
                continue;
            }
            SubstringPosition substringPosition = fichePertinence.getSubstringPosition(operandNumber);
            OccurrencesBuilder builder = builderByAlineaMap.get(alineaInteger);
            if (builder == null) {
                builder = new OccurrencesBuilder(acceptLength);
                builderByAlineaMap.put(alineaInteger, builder);
            }
            builder.add(operandNumber, substringPosition);
        }
        int size = builderByAlineaMap.size();
        InAlineaOccurrences[] occurrencesArray = new InAlineaOccurrences[size];
        int p = 0;
        for (Map.Entry<Integer, OccurrencesBuilder> entry : builderByAlineaMap.entrySet()) {
            occurrencesArray[p] = entry.getValue().toInAlineaOccurrences(entry.getKey());
            p++;
        }
        return occurrencesArray;
    }

    public static InLangOccurrences[] getInLangOccurrencesArray(MotclePertinence motclePertinence, int[] acceptOperandArray) {
        int langCount = motclePertinence.getLangOccurrenceCount();
        OccurrencesBuilder[] builderArray = new OccurrencesBuilder[langCount];
        Lang[] langArray = new Lang[langCount];
        int acceptLength = acceptOperandArray.length;
        for (int i = 0; i < langCount; i++) {
            MotclePertinence.LangOccurrence langOccurrence = motclePertinence.getLangOccurrence(i);
            langArray[i] = langOccurrence.getLang();
            for (int j = 0; j < acceptLength; j++) {
                int operandNumber = acceptOperandArray[j];
                SubstringPosition substringPosition = langOccurrence.getSubstringPosition(operandNumber);
                if (substringPosition == null) {
                    continue;
                }
                OccurrencesBuilder builder = builderArray[i];
                if (builder == null) {
                    builder = new OccurrencesBuilder(acceptLength);
                    builderArray[i] = builder;
                }
                builder.add(operandNumber, substringPosition);
            }
        }
        List<InLangOccurrences> list = new ArrayList<InLangOccurrences>(langCount);
        for (int i = 0; i < langCount; i++) {
            Lang lang = langArray[i];
            OccurrencesBuilder builder = builderArray[i];
            if (builder != null) {
                list.add(builder.toInLangOccurrences(lang));
            }
        }
        return list.toArray(new InLangOccurrences[list.size()]);
    }

    public static SortedSet<PertinencePack> addOriginScore(SortedSet<PertinencePack> set, float originPonderation, DataAccess dataAccess) {
        SortedSet<PertinencePack> resultSet = new TreeSet<PertinencePack>();
        Map<Integer, Integer> avanceMap = new HashMap<Integer, Integer>();
        for (PertinencePack pertinencePack : set) {
            int codecorpus = pertinencePack.getFichePertinence().getCodecorpus();
            int codebase = dataAccess.getCorpusDataByCode(codecorpus).getCodeBase();
            Integer avance = avanceMap.get(codebase);
            float originScore = 0;
            if (avance == null) {
                originScore = originPonderation;
                avanceMap.put(codebase, 128); //la seconde valeur sera à deux tiers 128 / 192
            } else {
                int newAvance = avance;
                if (newAvance > 1) {
                    originScore = originPonderation * newAvance / 192;
                    newAvance = newAvance / 2;
                    avanceMap.put(codebase, newAvance);
                }
            }
            if (originScore != 0) {
                resultSet.add(pertinencePack.derive(originScore));
            } else {
                resultSet.add(pertinencePack);
            }
        }
        return resultSet;
    }

    public static float getDateScore(FuzzyDate date, int currentYear) {
        if (date == null) {
            return 0;
        }
        int year = date.getYear();
        float x = (float) Math.abs(year - currentYear);
        return 1 - (x / (x + 6));
    }

    public static float getOccurrenceScore(FichePertinence fichePertinence, int[] acceptOperandArray) {
        int length = acceptOperandArray.length;
        int bestAlinea = -1;
        int bestBegin = -1;
        int note = 0;
        for (int i = 0; i < length; i++) {
            int operandNumber = acceptOperandArray[i];
            int alineaInteger = fichePertinence.getAlineaInteger(operandNumber);
            int motcleNote = convertMotcleNote(fichePertinence.getMotcleScore(operandNumber));
            if (alineaInteger > 0) {
                if ((bestAlinea == -1) || (alineaInteger < bestAlinea)) {
                    bestAlinea = alineaInteger;
                    bestBegin = fichePertinence.getSubstringPosition(operandNumber).getBeginIndex();
                } else if (alineaInteger == bestAlinea) {
                    if (bestBegin == -1) {
                        bestBegin = fichePertinence.getSubstringPosition(operandNumber).getBeginIndex();
                    } else {
                        bestBegin = Math.min(bestBegin, fichePertinence.getSubstringPosition(operandNumber).getBeginIndex());
                    }
                }
                note = note + getOperandNote(alineaInteger, motcleNote);
            } else if (motcleNote > 0) {
                note = note + 35 + motcleNote;
                if ((bestAlinea == -1) || (bestAlinea > AlineaInteger.COMPLEMENT_START)) {
                    bestAlinea = AlineaInteger.COMPLEMENT_START;
                    bestBegin = 10;
                }
            }
        }
        if (bestAlinea > 0) {
            note = note + getOperandNote(bestAlinea, 0); //On double la note du meilleur champ
            note = note + getBestIndexPrime(bestBegin);
        }
        int notemaximale = 115 * (length + 1) + 4;
        return ((float) note) / notemaximale;
    }

    private static int getBestIndexPrime(int bestIndex) {
        if (bestIndex == 0) {
            return 4;
        } else if (bestIndex < 5) {
            return 3;
        } else if (bestIndex < 20) {
            return 2;
        } else if (bestIndex < 50) {
            return 1;
        } else {
            return 0;
        }
    }

    private static int convertMotcleNote(float floatNote) {
        return (int) (15 * floatNote);
    }

    private static int getOperandNote(int alineaInteger, int motcleNote) {
        if (alineaInteger == AlineaInteger.TITRE) {
            return 100 + motcleNote;
        }
        if (alineaInteger == AlineaInteger.SOUSTITRE) {
            return 80 + motcleNote;
        }
        int alineaCategory = AlineaInteger.getAlineaCategory(alineaInteger);
        switch (alineaCategory) {
            case AlineaInteger.PRIMARYATTRIBUTE_CATEGORY:
                return 60 + motcleNote;
            case AlineaInteger.COMPLEMENT_CATEGORY:
                if (motcleNote > 0) {
                    return 50 + motcleNote;
                }
                return 40;
            case AlineaInteger.SECONDARYATTRIBUTE_CATEGORY:
                if (motcleNote > 0) {
                    return 40 + motcleNote;
                }
                return 20;
            default:
                return 0;
        }
    }


}
