/* ScrutariLib_Supermotcle - Copyright (c) 2010-2017 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package net.scrutari.supermotcle.tools;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.RandomAccess;
import java.util.Set;
import java.util.TreeSet;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.text.collation.CollationUnit;
import net.scrutari.data.BaseData;
import net.scrutari.db.api.DataAccess;
import net.scrutari.db.api.codes.ElementCodes;
import net.scrutari.db.api.codes.IndexationCodes;
import net.scrutari.db.tools.codes.CodesUtils;
import net.scrutari.supermotcle.api.Supermotcle;


/**
 *
 * @author Vincent Calame
 */
public class SupermotcleBuilder {

    private final static IndexationCounterComparator INDEXATION_COMPARATOR = new IndexationCounterComparator();
    private final Lang lang;
    private final int code;
    private final List<Integer> codemotcleList = new ArrayList<Integer>();
    private final List<Integer> codethesaurusList = new ArrayList<Integer>();
    private final DataAccess dataAccess;
    private final List<IndexationCounter> indexationCounterList = new ArrayList<IndexationCounter>();
    private CollationUnit collationUnit;

    public SupermotcleBuilder(DataAccess dataAccess, Lang lang, int code, int codethesaurus, int codemotcle, CollationUnit collationUnit) {
        this.dataAccess = dataAccess;
        this.lang = lang;
        this.code = code;
        this.collationUnit = collationUnit;
        codemotcleList.add(codemotcle);
        codethesaurusList.add(codethesaurus);
        int baseCount = dataAccess.getBaseCount();
        for (int i = 0; i < baseCount; i++) {
            BaseData baseData = dataAccess.getBaseDataByIndex(i);
            int codebase = baseData.getCodeBase();
            int corpusCount = baseData.getCorpusCount();
            int[] codecorpusArray = new int[corpusCount];
            for (int j = 0; j < corpusCount; j++) {
                codecorpusArray[j] = baseData.getCodeCorpus(j);
            }
            IndexationCounter indexationCounter = new IndexationCounter(codebase, codecorpusArray);
            indexationCounterList.add(indexationCounter);
        }
        countIndexation(codemotcle);
    }

    public void addMotcle(int codethesaurus, int codemotcle, CollationUnit newCollationUnit) { //On compare pour voir quelle est la meilleur « collation »
        String oldSourceString = collationUnit.getSourceString();
        String newSourceString = newCollationUnit.getSourceString();
        boolean replace = false;
        int newLength = newSourceString.length();
        int oldLength = oldSourceString.length();
        if (newLength < oldLength) {
            replace = true;
        } else if (newLength == oldLength) {
            if (newSourceString.compareTo(oldSourceString) > 0) {
                replace = true;
            }
        }
        if (replace) {
            collationUnit = newCollationUnit;
        }
        codemotcleList.add(codemotcle);
        codethesaurusList.add(codethesaurus);
        countIndexation(codemotcle);
    }

    public Supermotcle toSupermotcle() {
        ElementCodes motcleCodes = CodesUtils.toElementCodes(codemotcleList, codethesaurusList);
        Set<IndexationCounter> resultSet = new TreeSet<IndexationCounter>(INDEXATION_COMPARATOR);
        int globalIndexationCount = 0;
        for (IndexationCounter indexationCounter : indexationCounterList) {
            if (indexationCounter.indexationCount > 0) {
                globalIndexationCount = globalIndexationCount + indexationCounter.indexationCount;
                resultSet.add(indexationCounter);
            }
        }
        List<Supermotcle.IndexationByBase> indexationByBaseList = new IndexationByBaseList(resultSet.toArray(new Supermotcle.IndexationByBase[resultSet.size()]));
        return new InternalSupermotcle(lang, code, collationUnit, motcleCodes, globalIndexationCount, indexationByBaseList);
    }

    private void countIndexation(int codemotcle) {
        int baseCount = indexationCounterList.size();
        for (int i = 0; i < baseCount; i++) {
            IndexationCounter indexationCounter = indexationCounterList.get(i);
            indexationCounter.checkMotcle(codemotcle);
        }
    }


    private class IndexationCounter implements Supermotcle.IndexationByBase {

        private final int codebase;
        private final int[] codecorpusArray;
        private int indexationCount;

        private IndexationCounter(int codebase, int[] codecorpusArray) {
            this.codebase = codebase;
            this.codecorpusArray = codecorpusArray;
            this.indexationCount = 0;
        }

        private void checkMotcle(int codemotcle) {
            int length = codecorpusArray.length;
            for (int i = 0; i < length; i++) {
                IndexationCodes indexationCodes = dataAccess.getIndexationCodes(codemotcle, codecorpusArray[i]);
                int ficheCount = indexationCodes.getCodeCount();
                indexationCount = indexationCount + ficheCount;
            }
        }

        @Override
        public int getBaseCode() {
            return codebase;
        }

        @Override
        public int getIndexationCount() {
            return indexationCount;
        }

    }


    private static class InternalSupermotcle implements Supermotcle {

        private final Lang lang;
        private final int code;
        private final CollationUnit collationUnit;
        private final ElementCodes motcleCodes;
        private final int indexationCount;
        private final List<Supermotcle.IndexationByBase> indexationByBaseList;

        private InternalSupermotcle(Lang lang, int code, CollationUnit collationUnit, ElementCodes motcleCodes, int indexationCount, List<Supermotcle.IndexationByBase> indexationByBaseList) {
            this.lang = lang;
            this.code = code;
            this.collationUnit = collationUnit;
            this.motcleCodes = motcleCodes;
            this.indexationCount = indexationCount;
            this.indexationByBaseList = indexationByBaseList;
        }

        @Override
        public Lang getLang() {
            return lang;
        }

        @Override
        public int getCode() {
            return code;
        }

        @Override
        public CollationUnit getCollationUnit() {
            return collationUnit;
        }

        @Override
        public ElementCodes getMotcleCodes() {
            return motcleCodes;
        }

        @Override
        public int getIndexationCount() {
            return indexationCount;
        }

        @Override
        public List<IndexationByBase> getIndexationByBaseList() {
            return indexationByBaseList;
        }

    }


    private static class IndexationCounterComparator implements Comparator<IndexationCounter> {

        private IndexationCounterComparator() {
        }

        @Override
        public int compare(IndexationCounter s1, IndexationCounter s2) {
            int ic1 = s1.indexationCount;
            int ic2 = s2.indexationCount;
            if (ic1 > ic2) {
                return -1;
            }
            if (ic1 < ic2) {
                return 1;
            }
            int c1 = s1.codebase;
            int c2 = s2.codebase;
            if (c1 < c2) {
                return -1;
            }
            if (c2 > c1) {
                return 1;
            }
            return 0;
        }

    }


    private static class IndexationByBaseList extends AbstractList<Supermotcle.IndexationByBase> implements RandomAccess {

        private final Supermotcle.IndexationByBase[] array;

        private IndexationByBaseList(Supermotcle.IndexationByBase[] array) {
            this.array = array;
        }

        @Override
        public int size() {
            return array.length;
        }

        @Override
        public Supermotcle.IndexationByBase get(int index) {
            return array[index];
        }

    }

}
