/* SctServer - Copyright (c) 2015-2017 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.sctserver.json;

import fr.exemole.sctserver.request.FicheFieldsFilter;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import net.mapeadores.util.attr.Attribute;
import net.mapeadores.util.attr.AttributeDef;
import net.mapeadores.util.attr.AttributeHolder;
import net.mapeadores.util.attr.AttributeKey;
import net.mapeadores.util.json.JSONWriter;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.primitives.FuzzyDate;
import net.scrutari.data.FicheData;
import net.scrutari.data.ScrutariAttributes;
import net.scrutari.datauri.FicheURI;
import net.scrutari.db.api.AttributeKeyCoder;
import net.scrutari.db.tools.util.BaseCheck;
import net.scrutari.lexie.AlineaInteger;
import net.scrutari.searchengine.api.result.FicheSearchResultInfo;
import net.scrutari.searchengine.api.result.InAlineaOccurrences;
import net.scrutari.searchengine.api.result.SearchTokenOccurrence;


/**
 *
 * @author Vincent Calame
 */
public class FicheDataJson {

    private final static String[] RESERVED_ARRAY = {"codefiche", "codecorpus", "codebase",
        "basename", "corpusname", "ficheid", "titre", "soustitre", "href", "lang", "annee", "date",
        "ficheicon", "icon", "lat", "lon", "mtitre", "msoustitre", "score"};

    private FicheDataJson() {

    }

    public static boolean isReservedPropertyName(String name) {
        for (String reserved : RESERVED_ARRAY) {
            if (reserved.equals(name)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Note : comme FicheFieldsFilter est susceptible de proposer des noms de
     * propriétés différentes des noms standard, il faudra être vigileant en cas
     * d'ajout d'autres propriétés à l'objet JSon après l'appel de cette méthode
     * object. On rajoutera sans risque de conflit des noms de propriétés avec «
     * : » ou une majuscule (par exemple, motcleArray, ..) ou un tiret « - ».
     */
    public static void object(JSONWriter jsonWriter, FicheData ficheData, FicheFieldsFilter ficheFieldsFilter, BaseCheck baseCheck, FicheSearchResultInfo info, AttributeKeyCoder attributeKeyCoder) throws IOException {
        jsonWriter.key("codefiche");
        jsonWriter.value(ficheData.getCodeFiche());
        if (ficheFieldsFilter.isWithCodeCorpus()) {
            jsonWriter.key("codecorpus");
            jsonWriter.value(ficheData.getCodeCorpus());
        }
        if (ficheFieldsFilter.isWithCodeBase()) {
            jsonWriter.key("codebase");
            jsonWriter.value(ficheData.getCodeBase());
        }
        if (ficheFieldsFilter.isWithBaseName()) {
            FicheURI ficheURI = ficheData.getFicheURI();
            jsonWriter.key("basename");
            jsonWriter.value(ficheURI.getCorpusURI().getBaseURI().getBaseName());
        }
        if (ficheFieldsFilter.isWithCorpusName()) {
            FicheURI ficheURI = ficheData.getFicheURI();
            jsonWriter.key("corpusname");
            jsonWriter.value(ficheURI.getCorpusURI().getCorpusName());
        }
        if (ficheFieldsFilter.isWithFicheId()) {
            FicheURI ficheURI = ficheData.getFicheURI();
            jsonWriter.key("ficheid");
            jsonWriter.value(ficheURI.getFicheId());
        }
        String titrePropertyName = ficheFieldsFilter.getTitrePropertyName();
        if (titrePropertyName != null) {
            String titre = ficheData.getTitre();
            if (titre != null) {
                jsonWriter.key(titrePropertyName);
                jsonWriter.value(ficheData.getTitre());
            }
        }
        String soustitrePropertyName = ficheFieldsFilter.getSoustitrePropertyName();
        if (soustitrePropertyName != null) {
            String soustitre = ficheData.getSoustitre();
            if (soustitre != null) {
                jsonWriter.key(soustitrePropertyName);
                jsonWriter.value(soustitre);
            }
        }
        String hrefPropertyName = ficheFieldsFilter.getHrefPropertyName();
        if (hrefPropertyName != null) {
            String href = ficheData.getHref(baseCheck.getCheckedLangUi());
            if (href != null) {
                jsonWriter.key(hrefPropertyName);
                jsonWriter.value(href);
            }
        }
        if (ficheFieldsFilter.isWithLang()) {
            Lang lang = ficheData.getLang();
            if (!lang.equals(FicheData.UNDETERMINED_LANG)) {
                jsonWriter.key("lang");
                jsonWriter.value(lang.toString());
            }
        }
        String anneePropertyName = ficheFieldsFilter.getAnneePropertyName();
        if (anneePropertyName != null) {
            FuzzyDate date = ficheData.getDate();
            if (date != null) {
                jsonWriter.key(anneePropertyName);
                jsonWriter.value(date.getYear());
            }
        }
        if (ficheFieldsFilter.isWithDate()) {
            FuzzyDate date = ficheData.getDate();
            if (date != null) {
                jsonWriter.key("date");
                jsonWriter.value(date.toString());
            }
        }
        if (ficheFieldsFilter.isWithFicheIcon()) {
            String ficheIcon = ficheData.getFicheIcon();
            if (ficheIcon != null) {
                jsonWriter.key("ficheicon");
                jsonWriter.value(ficheIcon);
            }
        }
        if (ficheFieldsFilter.isWithIcon()) {
            String icon = ficheData.getFicheIcon();
            if (icon == null) {
                icon = baseCheck.getBaseIcon();
            }
            if (icon != null) {
                jsonWriter.key("icon");
                jsonWriter.value(icon);
            }
        }
        if (ficheFieldsFilter.isWithGeo() && ficheData.isWithGeo()) {
            jsonWriter.key("lat");
            jsonWriter.value(ficheData.getLatitude().toString());
            jsonWriter.key("lon");
            jsonWriter.value(ficheData.getLongitude().toString());
        }
        if (ficheFieldsFilter.isWithComplements()) {
            int maxNumber = ficheData.getComplementMaxNumber();
            if (maxNumber > 0) {
                boolean inclus = false;
                for (int num = 1; num <= maxNumber; num++) {
                    String comp = ficheData.getComplementByNumber(num);
                    if (comp == null) {
                        continue;
                    }
                    if (!inclus) {
                        jsonWriter.key("complementArray");
                        jsonWriter.array();
                        inclus = true;
                    }
                    jsonWriter.object();
                    jsonWriter.key("num");
                    jsonWriter.value(num);
                    jsonWriter.key("comp");
                    jsonWriter.value(comp);
                    jsonWriter.endObject();
                }
                if (inclus) {
                    jsonWriter.endArray();
                }
            }
        }
        if (ficheFieldsFilter.withAttributes()) {
            AttributeHolder attributeHolder = ficheData.getAttributHolder();
            int attributeCount = attributeHolder.getAttributeCount();
            for (int i = 0; i < attributeCount; i++) {
                Attribute attribute = attributeHolder.getAttributeByIndex(i);
                AttributeKey attributeKey = attribute.getAttributeKey();
                String propertyName = ficheFieldsFilter.getAttributePropertyName(attributeKey);
                if (propertyName != null) {
                    jsonWriter.key(propertyName);
                    AttributeDef attributeDef = attributeKeyCoder.getAttributeDef(attributeKey);
                    jsonWriter.value(attributeToString(attribute, attributeDef.getFormatType()));
                }
            }
        }
        if (info == null) {
            return;
        }
        Map<Integer, InAlineaOccurrences> map = new HashMap<Integer, InAlineaOccurrences>();
        Set<Integer> alineaRootCodeSet = null;
        List<InAlineaOccurrences> occurrencesList = info.getInAlineaOccurrencesList();
        int occurrencesLength = occurrencesList.size();
        for (int i = 0; i < occurrencesLength; i++) {
            InAlineaOccurrences occurrences = occurrencesList.get(i);
            int alineaInteger = occurrences.getAlineaInteger();
            map.put(alineaInteger, occurrences);
            if (AlineaInteger.isAttributeAlineaInteger(alineaInteger)) {
                int alineaRootCode = AlineaInteger.toAlineaRootCode(alineaInteger);
                if (alineaRootCodeSet == null) {
                    alineaRootCodeSet = new HashSet<Integer>();
                }
                alineaRootCodeSet.add(alineaRootCode);
            }
        }
        if (ficheFieldsFilter.isWithMTitre()) {
            String titre = ficheData.getTitre();
            if (titre != null) {
                jsonWriter.key("mtitre");
                JsonUtils.writeMLib(jsonWriter, titre, getList(map, AlineaInteger.TITRE));
            }
        }
        if (ficheFieldsFilter.isWithMSoustitre()) {
            String soustitre = ficheData.getSoustitre();
            if (soustitre != null) {
                jsonWriter.key("msoustitre");
                JsonUtils.writeMLib(jsonWriter, soustitre, getList(map, AlineaInteger.SOUSTITRE));
            }
        }
        if (ficheFieldsFilter.isWithMComplements()) {
            boolean all = ficheFieldsFilter.isAllMComplements();
            int maxNumber = ficheData.getComplementMaxNumber();
            if (maxNumber > 0) {
                boolean inclus = false;
                for (int num = 1; num <= maxNumber; num++) {
                    String comp = ficheData.getComplementByNumber(num);
                    if (comp == null) {
                        continue;
                    }
                    List<SearchTokenOccurrence> searchTokenOccurrenceList = getList(map, AlineaInteger.getComplementAlineaInteger(num));
                    if ((searchTokenOccurrenceList == null) && (!all)) {
                        continue;
                    }
                    if (!inclus) {
                        jsonWriter.key("mcomplementArray");
                        jsonWriter.array();
                        inclus = true;
                    }
                    jsonWriter.object();
                    jsonWriter.key("num");
                    jsonWriter.value(num);
                    jsonWriter.key("mcomp");
                    JsonUtils.writeMLib(jsonWriter, comp, searchTokenOccurrenceList);
                    jsonWriter.endObject();
                }
                if (inclus) {
                    jsonWriter.endArray();
                }
            }
        }
        short mAttributesFilter = ficheFieldsFilter.getMAttributesFilter();
        if (mAttributesFilter != FicheFieldsFilter.NONE_FILTER) {
            AttributeHolder attributeHolder = ficheData.getAttributHolder();
            int attributeCount = attributeHolder.getAttributeCount();
            boolean inclus = false;
            for (int i = 0; i < attributeCount; i++) {
                Attribute attribute = attributeHolder.getAttributeByIndex(i);
                AttributeKey attributeKey = attribute.getAttributeKey();
                int alineaRootCode = attributeKeyCoder.getAlineaRootCode(attributeKey);
                if (alineaRootCode < 1) {
                    continue;
                }
                AttributeDef attributeDef = attributeKeyCoder.getAttributeDef(attributeKey);
                boolean isNotConcerned = ((alineaRootCodeSet == null) || (!alineaRootCodeSet.contains(alineaRootCode)));
                boolean isBlock = (attributeDef.getFormatType().equals(ScrutariAttributes.BLOCK_TYPE));
                if ((isBlock) && (isNotConcerned)) {
                    continue;
                }
                boolean stop = false;
                switch (mAttributesFilter) {
                    case FicheFieldsFilter.CONCERNED_FILTER:
                        if (isNotConcerned) {
                            stop = true;
                        }
                        break;
                    case FicheFieldsFilter.PRIMARY_FILTER:
                        if (AlineaInteger.getAttributeCategory(alineaRootCode) == AlineaInteger.SECONDARYATTRIBUTE_CATEGORY) {
                            if (isNotConcerned) {
                                stop = true;
                            }
                        }
                        break;
                }
                if (stop) {
                    continue;
                }
                if (!inclus) {
                    jsonWriter.key("mattrMap");
                    jsonWriter.object();
                    inclus = true;
                }
                jsonWriter.key(attributeKey.toString());
                int valCount = attribute.getValueCount();
                jsonWriter.array();
                for (int j = 0; j < valCount; j++) {
                    String value = attribute.getValue(j);
                    int alineaInteger = AlineaInteger.getValueAlineaInteger(alineaRootCode, j);
                    List<SearchTokenOccurrence> searchTokenOccurrenceList = getList(map, alineaInteger);
                    if ((searchTokenOccurrenceList != null) || (!isBlock)) {
                        JsonUtils.writeMLib(jsonWriter, value, searchTokenOccurrenceList);
                    }
                }
                jsonWriter.endArray();
            }
            if (inclus) {
                jsonWriter.endObject();
            }
        }
        if (ficheFieldsFilter.isWithScore()) {
            jsonWriter.key("score");
            int[] array = info.getScoreArray();
            int scoreLength = array.length;
            jsonWriter.array();
            for (int i = 0; i < scoreLength; i++) {
                jsonWriter.value(array[i]);
            }
            jsonWriter.endArray();
        }
    }

    private static List<SearchTokenOccurrence> getList(Map<Integer, InAlineaOccurrences> map, int value) {
        InAlineaOccurrences occurrences = map.get(value);
        if (occurrences == null) {
            return null;
        }
        return occurrences.getSearchTokenOccurrenceList();
    }

    private static String attributeToString(Attribute attribute, String formatType) {
        if (formatType.equals(ScrutariAttributes.UNIQUE_TYPE)) {
            return attribute.getValue(0);
        }
        StringBuilder buf = new StringBuilder();
        int count = attribute.getValueCount();
        if (formatType.equals(ScrutariAttributes.INLINE_TYPE)) {
            for (int i = 0; i < count; i++) {
                if (i > 0) {
                    buf.append(", ");
                }
                buf.append(attribute.getValue(i));
            }
        } else if (formatType.equals(ScrutariAttributes.LIST_TYPE)) {
            for (int i = 0; i < count; i++) {
                if (i > 0) {
                    buf.append("\n");
                }
                buf.append("- ");
                buf.append(attribute.getValue(i));
            }
        } else if (formatType.equals(ScrutariAttributes.BLOCK_TYPE)) {
            for (int i = 0; i < count; i++) {
                if (i > 0) {
                    buf.append("\n\n");
                }
                buf.append(attribute.getValue(i));
            }
        }
        return buf.toString();
    }

}
