/* SctServer - Copyright (c) 2009-2017 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.sctserver.tools.conf.factory;

import fr.exemole.sctserver.api.EngineMetadata;
import fr.exemole.sctserver.api.collect.ScrutariSourceDef;
import fr.exemole.sctserver.api.conf.ConfConstants;
import fr.exemole.sctserver.api.conf.ConfDirectory;
import fr.exemole.sctserver.api.conf.ConfFileInfo;
import fr.exemole.sctserver.api.conf.ConfIOException;
import fr.exemole.sctserver.api.feed.SctFeedMetadata;
import fr.exemole.sctserver.tools.DataValidatorBuilder;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Map;
import javax.xml.parsers.ParserConfigurationException;
import net.mapeadores.util.attr.AttributeDef;
import net.mapeadores.util.logging.CategorizedMessageEntry;
import net.scrutari.data.DataValidator;
import net.scrutari.datauri.ScrutariDataURI;
import net.scrutari.datauri.ScrutariDataURIChecker;
import net.scrutari.db.api.CodeMatch;
import net.scrutari.searchengine.api.options.CategoryDef;
import net.scrutari.searchengine.api.options.PertinencePonderation;


/**
 *
 * @author Vincent Calame
 */
public class ConfDirectoryFactory {


    private ConfDirectoryFactory() {
    }

    public static ConfDirectory fromDirectory(File confDir) {
        return new InternalConfDirectory(confDir);
    }


    private static class InternalConfDirectory implements ConfDirectory {

        private final File confDir;
        private final InternalConfFileInfo sourcesFileInfo = new InternalConfFileInfo();
        private final InternalConfFileInfo metadataFileInfo = new InternalConfFileInfo();
        private final InternalConfFileInfo categoriesFileInfo = new InternalConfFileInfo();
        private final InternalConfFileInfo searchLogFilterFileInfo = new InternalConfFileInfo();
        private final InternalConfFileInfo uriCodesFileInfo = new InternalConfFileInfo();
        private final InternalConfFileInfo attributesFileInfo = new InternalConfFileInfo();
        private final InternalConfFileInfo optionsFileInfo = new InternalConfFileInfo();
        private final InternalConfFileInfo validationFileInfo = new InternalConfFileInfo();
        private ScrutariSourceDef[] scrutariSourceDefArray = null;
        private EngineMetadata engineMetadata = null;
        private CategoryDef[] categoryDefArray = null;
        private PertinencePonderation globalPertinencePonderation;
        private CodeMatch[] initialCodeMatchArray = null;
        private SctFeedMetadata sctFeedMetadata;
        private Map<String, AttributeDef[]> attributeDefArrayMap;
        private DataValidator dataValidator = DataValidatorBuilder.NEUTRAL_DATAVALIDATOR;

        private InternalConfDirectory(File confDir) {
            this.confDir = confDir;
        }

        @Override
        public ScrutariSourceDef[] getScrutariSourceDefArray() {
            return scrutariSourceDefArray;
        }

        @Override
        public CategoryDef[] getCategoryDefArray() {
            return categoryDefArray;
        }

        @Override
        public PertinencePonderation getGlobalPertinencePonderation() {
            return globalPertinencePonderation;
        }

        @Override
        public CodeMatch[] getInitialCodeMatchArray(ScrutariDataURIChecker checker) {
            if (initialCodeMatchArray == null) {
                return null;
            }
            int length = initialCodeMatchArray.length;
            if (length == 0) {
                return null;
            }
            CodeMatch[] result = new CodeMatch[length];
            if (checker == null) {
                System.arraycopy(initialCodeMatchArray, 0, result, 0, length);
            } else {
                for (int i = 0; i < length; i++) {
                    CodeMatch codeMatch = initialCodeMatchArray[i];
                    ScrutariDataURI uri = checker.checkURI(codeMatch.getScrutariDataURI());
                    result[i] = new CodeMatch(codeMatch.getCode(), uri);
                }
            }
            return result;
        }

        @Override
        public SctFeedMetadata getSctFeedMetadata() {
            return sctFeedMetadata;
        }

        @Override
        public EngineMetadata getEngineMetadata() {
            return engineMetadata;
        }

        @Override
        public Map<String, AttributeDef[]> getAttributeDefArrayMap() {
            return attributeDefArrayMap;
        }

        @Override
        public DataValidator getDataValidator() {
            return dataValidator;
        }

        @Override
        public void reload() throws ConfIOException {
            clear();
            if (!confDir.exists()) {
                return;
            }
            try {
                if (!confDir.isDirectory()) {
                    throw new ConfIOException("confDirectory is not a directory : " + confDir.getPath());
                }
                for (File f : confDir.listFiles()) {
                    if (!f.isDirectory()) {
                        String name = f.getName();
                        if ((name.length() > 4) && (name.endsWith(".xml"))) {
                            String baseName = name.substring(0, name.length() - 4);
                            switch (baseName) {
                                case ConfConstants.SOURCES_NAME:
                                    initSources(f);
                                    break;
                                case ConfConstants.METADATA_NAME:
                                    initMetadata(f);
                                    break;
                                case ConfConstants.CATEGORIES_NAME:
                                    initCategories(f);
                                    break;
                                case ConfConstants.URI_CODES_NAME:
                                    initUriCodes(f);
                                    break;
                                case ConfConstants.ATTRIBUTES_NAME:
                                    initAttributes(f);
                                    break;
                                case ConfConstants.OPTIONS_NAME:
                                    initOptions(f);
                                    break;
                                case ConfConstants.VALIDATION_NAME:
                                    initValidation(f);
                                    break;
                            }
                        }
                    }
                }
            } catch (SecurityException | IOException | ParserConfigurationException e) {
                throw new ConfIOException(e);
            }
        }

        private void clear() {
            sourcesFileInfo.clear();
            metadataFileInfo.clear();
            categoriesFileInfo.clear();
            searchLogFilterFileInfo.clear();
            uriCodesFileInfo.clear();
            validationFileInfo.clear();
            scrutariSourceDefArray = null;
            categoryDefArray = null;
            initialCodeMatchArray = null;
            sctFeedMetadata = null;
            attributeDefArrayMap = null;
            dataValidator = DataValidatorBuilder.NEUTRAL_DATAVALIDATOR;
        }

        private void initSources(File f) throws IOException, ParserConfigurationException {
            sourcesFileInfo.setHere(true);
            SourcesConf sourcesConf = SourcesConf.load(f);
            sourcesFileInfo.setMessageList(sourcesConf.getMessageList());
            sourcesFileInfo.setProcessURI(sourcesConf.getProcessURI());
            scrutariSourceDefArray = sourcesConf.getScrutariSourceDefArray();
        }

        private void initMetadata(File f) throws IOException, ParserConfigurationException {
            metadataFileInfo.setHere(true);
            MetadataConf metadataConf = MetadataConf.load(f);
            metadataFileInfo.setMessageList(metadataConf.getMessageList());
            metadataFileInfo.setProcessURI(metadataConf.getProcessURI());
            engineMetadata = metadataConf.getEngineMetadata();
        }

        private void initCategories(File f) throws IOException, ParserConfigurationException {
            categoriesFileInfo.setHere(true);
            CategoriesConf categoriesConf = CategoriesConf.load(f);
            categoriesFileInfo.setMessageList(categoriesConf.getMessageList());
            categoriesFileInfo.setProcessURI(categoriesConf.getProcessURI());
            categoryDefArray = categoriesConf.getCategoryDefArray();
        }

        private void initUriCodes(File f) throws IOException, ParserConfigurationException {
            uriCodesFileInfo.setHere(true);
            UriCodesConf uriCodesConf = UriCodesConf.load(f);
            uriCodesFileInfo.setMessageList(uriCodesConf.getMessageList());
            uriCodesFileInfo.setProcessURI(uriCodesConf.getProcessURI());
            initialCodeMatchArray = uriCodesConf.getCodeMatchArray();
        }

        private void initAttributes(File f) throws IOException, ParserConfigurationException {
            attributesFileInfo.setHere(true);
            AttributesConf attributesConf = AttributesConf.load(f);
            attributesFileInfo.setMessageList(attributesConf.getMessageList());
            attributesFileInfo.setProcessURI(attributesConf.getProcessURI());
            attributeDefArrayMap = attributesConf.getAttributeDefArrayMap();
        }

        private void initOptions(File f) throws IOException, ParserConfigurationException {
            optionsFileInfo.setHere(true);
            OptionsConf optionsConf = OptionsConf.load(f);
            optionsFileInfo.setMessageList(optionsConf.getMessageList());
            optionsFileInfo.setProcessURI(optionsConf.getProcessURI());
            globalPertinencePonderation = optionsConf.getGlobalPertinencePonderation();
        }

        private void initValidation(File f) throws IOException, ParserConfigurationException {
            validationFileInfo.setHere(true);
            ValidationConf validationConf = ValidationConf.load(f);
            validationFileInfo.setMessageList(validationConf.getMessageList());
            validationFileInfo.setProcessURI(validationConf.getProcessURI());
            dataValidator = validationConf.getDataValidator();
        }

        @Override
        public ConfFileInfo getConfFileInfo(String name) {
            switch (name) {
                case ConfConstants.SOURCES_NAME:
                    return sourcesFileInfo;
                case ConfConstants.METADATA_NAME:
                    return metadataFileInfo;
                case ConfConstants.CATEGORIES_NAME:
                    return categoriesFileInfo;
                case ConfConstants.URI_CODES_NAME:
                    return uriCodesFileInfo;
                case ConfConstants.ATTRIBUTES_NAME:
                    return attributesFileInfo;
                case ConfConstants.OPTIONS_NAME:
                    return optionsFileInfo;
                case ConfConstants.VALIDATION_NAME:
                    return validationFileInfo;
                default:
                    throw new IllegalArgumentException("Unknown name : " + name);
            }
        }

    }


    private static class InternalConfFileInfo implements ConfFileInfo {

        private boolean here = false;
        private String processURI;
        private List<CategorizedMessageEntry> messageList;

        private InternalConfFileInfo() {
        }

        private void clear() {
            this.here = false;
            this.processURI = null;
            this.messageList = null;
        }

        @Override
        public boolean isHere() {
            return here;
        }

        @Override
        public String getProcessURI() {
            return processURI;
        }

        @Override
        public List<CategorizedMessageEntry> getMessageList() {
            return messageList;
        }

        private void setHere(boolean here) {
            this.here = here;
        }

        private void setMessageList(List<CategorizedMessageEntry> messageList) {
            this.messageList = messageList;
        }

        private void setProcessURI(String processURI) {
            this.processURI = processURI;
        }

    }


}
