/* SctServer_Directory - Copyright (c) 2012-2014 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.sctserver.directory.cache;

import fr.exemole.sctserver.api.storage.SctStorageException;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.HashMap;
import net.mapeadores.util.primitives.io.DataOutputPrimitivesWriter;
import net.scrutari.data.BaseData;
import net.scrutari.data.CorpusData;
import net.scrutari.data.FicheData;
import net.scrutari.data.IndexationData;
import net.scrutari.data.MotcleData;
import net.scrutari.data.ThesaurusData;
import net.scrutari.db.api.caches.DataCacheReaderFactory;
import net.scrutari.db.api.caches.DataCacheWriter;


/**
 *
 * @author Vincent Calame
 */
public class DirectoryDataCacheWriter implements DataCacheWriter {

    private final HashMap<Integer, Long> positionMap = new HashMap<Integer, Long>();
    private final DataWriteInfo[] dataWriteInfoArray;

    public DirectoryDataCacheWriter(File cacheDirectory) {
        dataWriteInfoArray = new DataWriteInfo[CacheUtils.LENGTH];
        dataWriteInfoArray[CacheUtils.BASE_INDEX] = initDataWriteInfo(cacheDirectory, "base");
        dataWriteInfoArray[CacheUtils.CORPUS_INDEX] = initDataWriteInfo(cacheDirectory, "corpus");
        dataWriteInfoArray[CacheUtils.THESAURUS_INDEX] = initDataWriteInfo(cacheDirectory, "thesaurus");
        dataWriteInfoArray[CacheUtils.FICHE_INDEX] = initDataWriteInfo(cacheDirectory, "fiche");
        dataWriteInfoArray[CacheUtils.MOTCLE_INDEX] = initDataWriteInfo(cacheDirectory, "motcle");
        dataWriteInfoArray[CacheUtils.INDEXATION_INDEX] = initDataWriteInfo(cacheDirectory, "indexation");
    }

    @Override
    public void cacheBaseData(BaseData baseData) {
        DataWriteInfo dataWriteInfo = dataWriteInfoArray[CacheUtils.BASE_INDEX];
        try {
            BaseData.toPrimitives(baseData, dataWriteInfo.dataPrimitivesWriter);
        } catch (IOException ioe) {
            throw new SctStorageException(ioe);
        }
    }

    @Override
    public void cacheCorpusData(CorpusData corpusData) {
        DataWriteInfo dataWriteInfo = dataWriteInfoArray[CacheUtils.CORPUS_INDEX];
        try {
            CorpusData.toPrimitives(corpusData, dataWriteInfo.dataPrimitivesWriter);
        } catch (IOException ioe) {
            throw new SctStorageException(ioe);
        }
    }

    @Override
    public void cacheThesaurusData(ThesaurusData thesaurusData) {
        DataWriteInfo dataWriteInfo = dataWriteInfoArray[CacheUtils.THESAURUS_INDEX];
        try {
            ThesaurusData.toPrimitives(thesaurusData, dataWriteInfo.dataPrimitivesWriter);
        } catch (IOException ioe) {
            throw new SctStorageException(ioe);
        }
    }

    @Override
    public void cacheFicheData(FicheData ficheData) {
        DataWriteInfo dataWriteInfo = dataWriteInfoArray[CacheUtils.FICHE_INDEX];
        try {
            long position = dataWriteInfo.dataRandomAccessFile.getFilePointer();
            int codefiche = ficheData.getCodeFiche();
            positionMap.put(codefiche, position);
            FicheData.toPrimitives(ficheData, dataWriteInfo.dataPrimitivesWriter);
        } catch (IOException ioe) {
            throw new SctStorageException(ioe);
        }
    }

    @Override
    public void cacheMotcleData(MotcleData motcleData) {
        DataWriteInfo dataWriteInfo = dataWriteInfoArray[CacheUtils.MOTCLE_INDEX];
        try {
            long position = dataWriteInfo.dataRandomAccessFile.getFilePointer();
            int codemotcle = motcleData.getCodeMotcle();
            positionMap.put(codemotcle, position);
            MotcleData.toPrimitives(motcleData, dataWriteInfo.dataPrimitivesWriter);
        } catch (IOException ioe) {
            throw new SctStorageException(ioe);
        }
    }

    @Override
    public void cacheIndexationData(IndexationData indexationData) {
        DataWriteInfo dataWriteInfo = dataWriteInfoArray[CacheUtils.INDEXATION_INDEX];
        try {
            IndexationData.toPrimitives(indexationData, dataWriteInfo.dataPrimitivesWriter);
        } catch (IOException ioe) {
            throw new SctStorageException(ioe);
        }
    }

    @Override
    public DataCacheReaderFactory endCacheWrite() {
        File[] dataFileArray = new File[CacheUtils.LENGTH];
        for (int i = 0; i < CacheUtils.LENGTH; i++) {
            DataWriteInfo dataFileWriter = dataWriteInfoArray[i];
            dataFileArray[i] = dataFileWriter.dataFile;
            dataFileWriter.close();
            dataWriteInfoArray[i] = null;
        }
        return new DirectoryDataCacheReaderFactory(positionMap, dataFileArray);
    }

    private DataWriteInfo initDataWriteInfo(File cacheDirectory, String name) {
        File f = new File(cacheDirectory, "data-" + name + ".primitives");
        if (f.exists()) {
            f.delete();
        }
        RandomAccessFile dataRandomAccessFile;
        try {
            dataRandomAccessFile = new RandomAccessFile(f, "rw");
        } catch (IOException ioe) {
            throw new SctStorageException(ioe);
        }
        return new DataWriteInfo(f, dataRandomAccessFile);
    }


    private static class DataWriteInfo {

        private final File dataFile;
        private final RandomAccessFile dataRandomAccessFile;
        private final DataOutputPrimitivesWriter dataPrimitivesWriter;

        private DataWriteInfo(File dataFile, RandomAccessFile dataRandomAccessFile) {
            this.dataFile = dataFile;
            this.dataRandomAccessFile = dataRandomAccessFile;
            this.dataPrimitivesWriter = new DataOutputPrimitivesWriter(dataRandomAccessFile);
        }

        private void close() {
            try {
                dataRandomAccessFile.close();
            } catch (IOException ioe) {
                throw new SctStorageException(ioe);
            }
        }

    }


}
