/* UtilLib - Copyright (c) 2006-2017 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package net.mapeadores.util.html;

import java.io.PrintWriter;
import java.io.Writer;
import java.util.Map;
import net.mapeadores.util.io.AppendableWriter;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.localisation.Message;
import net.mapeadores.util.localisation.MessageLocalisation;
import net.mapeadores.util.localisation.MessageLocalisationPack;


/**
 *
 * @author Vincent Calame
 */
public class HtmlPrinter {

    protected PrintWriter pw;
    private final MessageLocalisationPack messageLocalisationPack = new MessageLocalisationPack();
    private final MessageLocalisation messageLocalisation = messageLocalisationPack.getMessageLocalisationInterface();
    protected boolean breakLine = true;
    protected boolean withJavascript = false;

    public HtmlPrinter() {
    }

    public void addMessageLocalisation(MessageLocalisation messageLocalisation) {
        addMessageLocalisation(messageLocalisation, 1);
    }

    public void addMessageLocalisation(MessageLocalisation messageLocalisation, int priority) {
        messageLocalisationPack.addMessageLocalisation(messageLocalisation, priority);
    }

    public MessageLocalisation getMessageLocalisation() {
        return messageLocalisation;
    }

    public PrintWriter initPrinter(PrintWriter pw) {
        PrintWriter old = this.pw;
        this.pw = pw;
        return old;
    }

    public PrintWriter initPrinter(Appendable appendable) {
        PrintWriter old = this.pw;
        if (appendable instanceof PrintWriter) {
            this.pw = (PrintWriter) appendable;
        } else if (appendable instanceof Writer) {
            this.pw = new PrintWriter((Writer) appendable);
        } else {
            this.pw = new PrintWriter(new AppendableWriter(appendable));
        }
        return old;
    }

    public void setWithJavascript(boolean withJavascript) {
        this.withJavascript = withJavascript;
    }

    public boolean isWithJavascript() {
        return withJavascript;
    }

    public void flush() {
        pw.flush();
    }

    public PrintWriter getPrintWriter() {
        return pw;
    }


    public HtmlPrinter append(TrustedHtml trustedHtml) {
        pw.append(trustedHtml);
        return this;
    }

    public HtmlPrinter escape(char c) {
        return escape(c, false);
    }

    public HtmlPrinter escape(char carac, boolean preformatted) {
        switch (carac) {
            case '&':
                pw.print("&amp;");
                break;
            case '"':
                pw.print("&quot;");
                break;
            case '<':
                pw.print("&lt;");
                break;
            case '>':
                pw.print("&gt;");
                break;
            case '\'':
                pw.print("&#x27;");
                break;
            case '\\':
                pw.print("&#x5C;");
                break;
            case '`':
                pw.print("&#x60;");
                break;
            default:
                if ((!preformatted) && (carac < 32)) {
                    pw.print(' ');
                } else {
                    pw.print(carac);
                }
        }
        return this;
    }

    public HtmlPrinter append(int i) {
        pw.print(i);
        return this;
    }

    public HtmlPrinter append(long l) {
        pw.print(l);
        return this;
    }

    public HtmlPrinter colon() {
        localize("_ label.global.colon");
        space();
        return this;
    }

    public HtmlPrinter dash() {
        pw.print(" – ");
        return this;
    }

    public HtmlPrinter space() {
        pw.print(' ');
        return this;
    }

    public HtmlPrinter newLine() {
        pw.print('\n');
        return this;
    }

    public HtmlPrinter nonBreakableSpace() {
        pw.print("&nbsp;");
        return this;
    }

    public HtmlPrinter doublespace() {
        pw.print("&nbsp; ");
        return this;
    }

    public HtmlPrinter escape(Object obj) {
        if (obj == null) {
            pw.print("null");
            return this;
        }
        return escape(obj.toString());
    }

    public HtmlPrinter escape(CharSequence s) {
        return escape(s, false);
    }

    public HtmlPrinter escape(CharSequence s, boolean preformatted) {
        if (s == null) {
            pw.print("null");
            return this;
        }
        int length = s.length();
        for (int i = 0; i < length; i++) {
            escape(s.charAt(i), preformatted);
        }
        return this;
    }

    public HtmlPrinter breakLine() {
        if (breakLine) {
            pw.println("");
        }
        return this;
    }

    public HtmlPrinter scriptLineEnd() {
        pw.println(";");
        return this;
    }

    public HtmlPrinter scriptLiteral(CharSequence s) {
        if (s == null) {
            pw.print("null");
            return this;
        }
        pw.print('"');
        int length = s.length();
        for (int i = 0; i < length; i++) {
            char carac = s.charAt(i);
            switch (carac) {
                case '\n':
                    pw.print("\\n");
                    break;
                case '\t':
                    pw.print("\\t");
                    break;
                case '\r':
                    pw.print("\\r");
                    break;
                case '"':
                    pw.print("\\\"");
                    break;
                case '\'':
                    pw.print("\\'");
                    break;
                case '\\':
                    pw.print("\\\\");
                    break;
                case '<':
                    pw.print("\\u003c");
                    break;
                case '>':
                    pw.print("\\u003e");
                    break;
                case '&':
                    pw.print("\\u0026");
                    break;
                default:
                    if (carac < 32) {
                        pw.print(' ');
                    } else {
                        pw.print(carac);
                    }
            }
        }
        pw.print('"');
        return this;
    }


    public String getLocalization(String messageKey) {
        String loc = messageLocalisation.toString(messageKey);
        if (loc == null) {
            loc = "?" + messageKey + "?";
        }
        return loc;
    }

    public String getLocalization(Message message) {
        String loc = messageLocalisation.toString(message);
        if (loc == null) {
            loc = "?" + message.getMessageKey() + "?";
        }
        return loc;
    }

    public HtmlPrinter localize(String messageKey) {
        escape(getLocalization(messageKey));
        return this;
    }

    public HtmlPrinter localize(Message message) {
        escape(getLocalization(message));
        return this;
    }

    public HtmlPrinter subroutine(boolean b) {
        return this;
    }

    public HtmlPrinter subroutine(Object o) {
        return this;
    }

    public HtmlPrinter A(HtmlAttributes htmlAttributes) {
        openTag("a");
        if (htmlAttributes != null) {
            printAttribute("href", htmlAttributes.href());
            printAttribute("target", htmlAttributes.target());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _A() {
        endTag("a");
        return this;
    }

    public HtmlPrinter BODY() {
        startTag("body");
        breakLine();
        return this;
    }

    public HtmlPrinter BODY(HtmlAttributes htmlAttributes) {
        openTag("body");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _BODY() {
        endTag("body");
        breakLine();
        return this;
    }

    public HtmlPrinter BR() {
        startTag("br");
        return this;
    }

    public HtmlPrinter BUTTON(HtmlAttributes htmlAttributes) {
        openTag("button");
        printHTMLAttributes(htmlAttributes);
        printBooleanAttribute("disabled", htmlAttributes.disabled());
        printAttribute("type", htmlAttributes.type());
        closeTag();
        return this;
    }

    public HtmlPrinter _BUTTON() {
        endTag("button");
        return this;
    }

    public HtmlPrinter COL(HtmlAttributes htmlAttributes) {
        openTag("col");
        if (htmlAttributes != null) {
            printAttribute("width", htmlAttributes.width());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter CODE() {
        startTag("code");
        return this;
    }

    public HtmlPrinter CODE(HtmlAttributes htmlAttributes) {
        openTag("code");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _CODE() {
        endTag("code");
        return this;
    }

    public HtmlPrinter COMMENT() {
        pw.print("<!-- ");
        return this;
    }

    public HtmlPrinter _COMMENT() {
        pw.print(" -->");
        return this;
    }

    public HtmlPrinter DEL() {
        startTag("del");
        return this;
    }

    public HtmlPrinter DEL(HtmlAttributes htmlAttributes) {
        openTag("del");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _DEL() {
        endTag("del");
        return this;
    }

    public HtmlPrinter DIV() {
        startTag("div");
        return this;
    }

    public HtmlPrinter DIV(HtmlAttributes htmlAttributes) {
        openTag("div");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _DIV() {
        endTag("div");
        breakLine();
        return this;
    }

    public HtmlPrinter DOCTYPE_html5() {
        pw.println("<!DOCTYPE html>");
        return this;
    }

    public HtmlPrinter DOCTYPE_frameset() {
        pw.println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Frameset//EN\" \"http://www.w3.org/TR/html4/frameset.dtd\">");
        return this;
    }

    public HtmlPrinter EM() {
        startTag("em");
        return this;
    }

    public HtmlPrinter EM(HtmlAttributes htmlAttributes) {
        openTag("em");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _EM() {
        endTag("em");
        return this;
    }

    public HtmlPrinter FIELDSET() {
        startTag("fieldset");
        return this;
    }

    public HtmlPrinter FIELDSET(HtmlAttributes htmlAttributes) {
        openTag("fieldset");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _FIELDSET() {
        endTag("fieldset");
        return this;
    }

    public HtmlPrinter FORM(String method, String action, String target, String enctype) {
        openTag("form");
        printAttribute("action", action);
        printAttribute("method", method);
        printAttribute("target", target);
        printAttribute("enctype", enctype);
        closeTag();
        return this;
    }

    public HtmlPrinter FORM(HtmlAttributes htmlAttributes) {
        openTag("form");
        if (htmlAttributes != null) {
            printAttribute("action", htmlAttributes.action());
            printAttribute("method", htmlAttributes.method());
            printAttribute("target", htmlAttributes.target());
            printAttribute("enctype", htmlAttributes.enctype());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter FORM_get(String action) {
        return FORM(HtmlConstants.GET_METHOD, action, null, null);
    }

    public HtmlPrinter FORM_get(String action, String target) {
        return FORM(HtmlConstants.GET_METHOD, action, target, null);
    }

    public HtmlPrinter FORM_get(HtmlAttributes htmlAttributes) {
        return FORM(htmlAttributes.method(HtmlConstants.GET_METHOD));
    }

    public HtmlPrinter FORM_post(String action) {
        return FORM(HtmlConstants.POST_METHOD, action, null, null);
    }

    public HtmlPrinter FORM_post(String action, String target) {
        return FORM(HtmlConstants.POST_METHOD, action, target, null);
    }

    public HtmlPrinter FORM_post(String action, String target, String enctype) {
        return FORM(HtmlConstants.POST_METHOD, action, target, enctype);
    }

    public HtmlPrinter FORM_post(HtmlAttributes htmlAttributes) {
        return FORM(htmlAttributes.method(HtmlConstants.POST_METHOD));
    }

    public HtmlPrinter _FORM() {
        endTag("form");
        return this;
    }

    public HtmlPrinter FRAME(String name, String src) {
        openTag("frame");
        printAttribute("name", name);
        printAttribute("src", src);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter FRAMESET_rows(String rows) {
        openTag("frameset");
        printAttribute("rows", rows);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter FRAMESET_cols(String cols) {
        openTag("frameset");
        printAttribute("cols", cols);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _FRAMESET() {
        endTag("frameset");
        breakLine();
        return this;
    }

    public HtmlPrinter H1() {
        startTag("h1");
        return this;
    }

    public HtmlPrinter H1(HtmlAttributes htmlAttributes) {
        openTag("h1");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _H1() {
        endTag("h1");
        breakLine();
        return this;
    }

    public HtmlPrinter H2() {
        startTag("h2");
        return this;
    }

    public HtmlPrinter H2(HtmlAttributes htmlAttributes) {
        openTag("h2");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _H2() {
        endTag("h2");
        breakLine();
        return this;
    }

    public HtmlPrinter H3() {
        startTag("h3");
        return this;
    }

    public HtmlPrinter H3(HtmlAttributes htmlAttributes) {
        openTag("h3");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _H3() {
        endTag("h3");
        breakLine();
        return this;
    }

    public HtmlPrinter H4() {
        startTag("h4");
        return this;
    }

    public HtmlPrinter H4(HtmlAttributes htmlAttributes) {
        openTag("h4");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _H4() {
        endTag("h4");
        breakLine();
        return this;
    }

    public HtmlPrinter H5() {
        startTag("h5");
        return this;
    }

    public HtmlPrinter H5(HtmlAttributes htmlAttributes) {
        openTag("h5");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _H5() {
        endTag("h5");
        breakLine();
        return this;
    }

    public HtmlPrinter H6() {
        startTag("h6");
        return this;
    }

    public HtmlPrinter H6(HtmlAttributes htmlAttributes) {
        openTag("h6");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _H6() {
        endTag("h6");
        breakLine();
        return this;
    }

    public HtmlPrinter HEAD() {
        startTag("head");
        breakLine();
        return this;
    }

    public HtmlPrinter HEAD(HtmlAttributes htmlAttributes) {
        openTag("head");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _HEAD() {
        endTag("head");
        breakLine();
        return this;
    }

    public HtmlPrinter HR() {
        startTag("hr");
        return this;
    }

    public HtmlPrinter HR(HtmlAttributes htmlAttributes) {
        openTag("hr");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter HTML() {
        startTag("html");
        breakLine();
        return this;
    }

    public HtmlPrinter HTML(Lang lang) {
        openTag("html");
        if (lang != null) {
            printAttribute("lang", lang.toString());
            if (lang.isRTLScript()) {
                printAttribute("dir", "rtl");
            }

        }
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _HTML() {
        endTag("html");
        return this;
    }

    public HtmlPrinter IFRAME(HtmlAttributes htmlAttributes) {
        openTag("iframe");
        if (htmlAttributes != null) {
            printAttribute("src", htmlAttributes.src());
            printAttribute("width", htmlAttributes.width());
            printAttribute("height", htmlAttributes.height());
            printAttribute("scrolling", htmlAttributes.scrolling());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _IFRAME() {
        endTag("iframe");
        breakLine();
        return this;
    }

    public HtmlPrinter IMG(HtmlAttributes htmlAttributes) {
        openTag("img");
        if (htmlAttributes != null) {
            printAttribute("src", htmlAttributes.src());
            printAttribute("alt", htmlAttributes.alt());
            printAttribute("width", htmlAttributes.width());
            printAttribute("height", htmlAttributes.height());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter INPUT(HtmlAttributes htmlAttributes, boolean preformatted) {
        openTag("input");
        if (htmlAttributes != null) {
            printAttribute("type", htmlAttributes.type());
            printPreformattedAttribute("value", htmlAttributes.value(), preformatted);
            printAttribute("size", htmlAttributes.size());
            printBooleanAttribute("checked", htmlAttributes.checked());
            printBooleanAttribute("disabled", htmlAttributes.disabled());
            printBooleanAttribute("readonly", htmlAttributes.readonly());
            printBooleanAttribute("required", htmlAttributes.required());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter INPUT_button(HtmlAttributes htmlAttributes) {
        return INPUT(htmlAttributes.type(HtmlConstants.BUTTON_TYPE), false);
    }

    public HtmlPrinter INPUT_checkbox(String name, String value, boolean checked) {
        openTag("input");
        printAttribute("type", HtmlConstants.CHECKBOX_TYPE);
        printAttribute("name", name);
        printAttribute("value", value);
        printBooleanAttribute("checked", checked);
        closeTag();
        return this;
    }

    public HtmlPrinter INPUT_checkbox(HtmlAttributes htmlAttributes) {
        return INPUT(htmlAttributes.type(HtmlConstants.CHECKBOX_TYPE), false);
    }

    public HtmlPrinter INPUT_file(String name, int size) {
        openTag("input");
        printAttribute("type", HtmlConstants.FILE_TYPE);
        printAttribute("name", name);
        if (size > 0) {
            printAttribute("size", String.valueOf(size));
        }
        closeTag();
        return this;
    }

    public HtmlPrinter INPUT_file(HtmlAttributes htmlAttributes) {
        return INPUT(htmlAttributes.type(HtmlConstants.FILE_TYPE), false);
    }

    public HtmlPrinter INPUT_hidden(String name, String value) {
        openTag("input");
        printAttribute("type", HtmlConstants.HIDDEN_TYPE);
        printAttribute("name", name);
        printAttribute("value", value);
        closeTag();
        return this;
    }

    public HtmlPrinter INPUT_hidden(Map<String, String> hiddenParametersMap) {
        for (Map.Entry<String, String> entry : hiddenParametersMap.entrySet()) {
            INPUT_hidden(entry.getKey(), entry.getValue());
        }
        return this;
    }

    public HtmlPrinter INPUT_hidden(HtmlAttributes htmlAttributes) {
        return INPUT(htmlAttributes.type(HtmlConstants.HIDDEN_TYPE), false);
    }

    public HtmlPrinter INPUT_password(String name, int size) {
        openTag("input");
        printAttribute("type", HtmlConstants.PASSWORD_TYPE);
        printAttribute("name", name);
        if (size > 0) {
            printAttribute("size", String.valueOf(size));
        }
        closeTag();
        return this;
    }

    public HtmlPrinter INPUT_password(HtmlAttributes htmlAttributes) {
        return INPUT(htmlAttributes.type(HtmlConstants.PASSWORD_TYPE), false);
    }

    public HtmlPrinter INPUT_radio(String name, String value, boolean checked) {
        openTag("input");
        printAttribute("type", HtmlConstants.RADIO_TYPE);
        printAttribute("name", name);
        printAttribute("value", value);
        printBooleanAttribute("checked", checked);
        closeTag();
        return this;
    }

    public HtmlPrinter INPUT_radio(HtmlAttributes htmlAttributes) {
        return INPUT(htmlAttributes.type(HtmlConstants.RADIO_TYPE), false);
    }

    public HtmlPrinter INPUT_submit(String value) {
        openTag("input");
        printAttribute("type", HtmlConstants.SUBMIT_TYPE);
        printAttribute("value", value);
        closeTag();
        return this;
    }

    public HtmlPrinter INPUT_submit(HtmlAttributes htmlAttributes) {
        return INPUT(htmlAttributes.type(HtmlConstants.SUBMIT_TYPE), false);
    }

    public HtmlPrinter INPUT_text(String name, String value, int size) {
        openTag("input");
        printAttribute("type", HtmlConstants.TEXT_TYPE);
        printAttribute("name", name);
        printAttribute("value", value);
        if (size > 0) {
            printAttribute("size", String.valueOf(size));
        }
        closeTag();
        return this;
    }

    public HtmlPrinter INPUT_text(HtmlAttributes htmlAttributes) {
        return INPUT(htmlAttributes.type(HtmlConstants.TEXT_TYPE), false);
    }


    public HtmlPrinter LABEL() {
        startTag("label");
        return this;
    }

    public HtmlPrinter LABEL(HtmlAttributes htmlAttributes) {
        openTag("label");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _LABEL() {
        endTag("label");
        return this;
    }

    public HtmlPrinter LEGEND() {
        startTag("legend");
        return this;
    }

    public HtmlPrinter LEGEND(HtmlAttributes htmlAttributes) {
        openTag("legend");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _LEGEND() {
        endTag("legend");
        breakLine();
        return this;
    }

    public HtmlPrinter LI() {
        startTag("li");
        breakLine();
        return this;
    }

    public HtmlPrinter LI(HtmlAttributes htmlAttributes) {
        openTag("li");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _LI() {
        endTag("li");
        breakLine();
        return this;
    }

    public HtmlPrinter LINK(HtmlAttributes htmlAttributes) {
        openTag("link");
        if (htmlAttributes != null) {
            printAttribute("href", htmlAttributes.href());
            printAttribute("rel", htmlAttributes.rel());
            printAttribute("type", htmlAttributes.type());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter META(HtmlAttributes htmlAttributes) {
        openTag("meta");
        if (htmlAttributes != null) {
            printAttribute("http-equiv", htmlAttributes.httpEquiv());
            printAttribute("content", htmlAttributes.content());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter META_htmlcontenttype() {
        openTag("meta");
        printAttribute("http-equiv", "Content-Type");
        printAttribute("content", "text/html; charset=UTF-8");
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter NOSCRIPT() {
        startTag("noscript");
        return this;
    }

    public HtmlPrinter _NOSCRIPT() {
        endTag("noscript");
        return this;
    }

    public HtmlPrinter OPTGROUP(HtmlAttributes htmlAttributes) {
        openTag("optgroup");
        if (htmlAttributes != null) {
            printAttribute("label", htmlAttributes.label());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _OPTGROUP() {
        endTag("optgroup");
        breakLine();
        return this;
    }

    public HtmlPrinter OPTION(String value, boolean selected) {
        openTag("option");
        printAttribute("value", value);
        printBooleanAttribute("selected", selected);
        closeTag();
        return this;
    }

    public HtmlPrinter OPTION(HtmlAttributes htmlAttributes) {
        openTag("option");
        if (htmlAttributes != null) {
            printAttribute("value", htmlAttributes.value());
            printBooleanAttribute("selected", htmlAttributes.selected());
            printBooleanAttribute("disabled", htmlAttributes.disabled());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _OPTION() {
        endTag("option");
        breakLine();
        return this;
    }

    public HtmlPrinter P() {
        startTag("p");
        return this;
    }

    public HtmlPrinter P(HtmlAttributes htmlAttributes) {
        openTag("p");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _P() {
        endTag("p");
        breakLine();
        return this;
    }

    public HtmlPrinter PRE() {
        startTag("pre");
        return this;
    }

    public HtmlPrinter PRE(HtmlAttributes htmlAttributes) {
        openTag("pre");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _PRE() {
        endTag("pre");
        return this;
    }

    public HtmlPrinter SCRIPT() {
        startTag("script");
        return this;
    }

    public HtmlPrinter SCRIPT(HtmlAttributes htmlAttributes) {
        openTag("script");
        if (htmlAttributes != null) {
            printAttribute("type", htmlAttributes.type());
            printAttribute("src", htmlAttributes.src());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _SCRIPT() {
        endTag("script");
        breakLine();
        return this;
    }

    public HtmlPrinter SELECT(String name) {
        openTag("select");
        printAttribute("name", name);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter SELECT(HtmlAttributes htmlAttributes) {
        openTag("select");
        printHTMLAttributes(htmlAttributes);
        printBooleanAttribute("multiple", htmlAttributes.multiple());
        printAttribute("size", htmlAttributes.size());
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _SELECT() {
        endTag("select");
        return this;
    }


    public HtmlPrinter SMALL() {
        startTag("small");
        return this;
    }

    public HtmlPrinter _SMALL() {
        endTag("small");
        return this;
    }

    public HtmlPrinter SPAN(HtmlAttributes htmlAttributes) {
        openTag("span");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _SPAN() {
        endTag("span");
        return this;
    }

    public HtmlPrinter STRONG() {
        startTag("strong");
        return this;
    }

    public HtmlPrinter STRONG(HtmlAttributes htmlAttributes) {
        openTag("strong");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _STRONG() {
        endTag("strong");
        return this;
    }

    public HtmlPrinter TABLE() {
        startTag("table");
        breakLine();
        return this;
    }

    public HtmlPrinter TABLE(HtmlAttributes htmlAttributes) {
        openTag("table");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _TABLE() {
        endTag("table");
        breakLine();
        return this;
    }

    public HtmlPrinter TBODY() {
        startTag("tbody");
        breakLine();
        return this;
    }

    public HtmlPrinter TBODY(HtmlAttributes htmlAttributes) {
        openTag("tbody");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _TBODY() {
        endTag("tbody");
        breakLine();
        return this;
    }

    public HtmlPrinter TD() {
        startTag("td");
        return this;
    }

    public HtmlPrinter TD(HtmlAttributes htmlAttributes) {
        openTag("td");
        if (htmlAttributes != null) {
            int colSpan = htmlAttributes.colspan();
            if (colSpan > 1) {
                printAttribute("colspan", String.valueOf(colSpan));
            }
            int rowSpan = htmlAttributes.rowspan();
            if (rowSpan > 1) {
                printAttribute("rowspan", String.valueOf(rowSpan));
            }
            printAttribute("width", htmlAttributes.width());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _TD() {
        endTag("td");
        breakLine();
        return this;
    }

    public HtmlPrinter TEXTAREA(String name, int cols, int rows) {
        openTag("textarea");
        printAttribute("name", name);
        printAttribute("cols", String.valueOf(cols));
        printAttribute("rows", String.valueOf(rows));
        closeTag();
        return this;
    }

    public HtmlPrinter TEXTAREA(HtmlAttributes htmlAttributes) {
        openTag("textarea");
        int cols = 0;
        int rows = 0;
        if (htmlAttributes != null) {
            cols = htmlAttributes.cols();
            rows = htmlAttributes.rows();
        }
        if (cols < 1) {
            cols = 50;
        }
        if (rows < 1) {
            rows = 3;
        }
        printAttribute("cols", String.valueOf(cols));
        printAttribute("rows", String.valueOf(rows));
        printBooleanAttribute("disabled", htmlAttributes.disabled());
        printBooleanAttribute("readonly", htmlAttributes.readonly());
        printBooleanAttribute("required", htmlAttributes.required());
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _TEXTAREA() {
        endTag("textarea");
        return this;
    }

    public HtmlPrinter TFOOT() {
        startTag("tfoot");
        breakLine();
        return this;
    }

    public HtmlPrinter TFOOT(HtmlAttributes htmlAttributes) {
        openTag("tfoot");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _TFOOT() {
        endTag("tfoot");
        breakLine();
        return this;
    }

    public HtmlPrinter TH() {
        startTag("th");
        return this;
    }

    public HtmlPrinter TH(HtmlAttributes htmlAttributes) {
        openTag("th");
        if (htmlAttributes != null) {
            int colSpan = htmlAttributes.colspan();
            if (colSpan > 1) {
                printAttribute("colspan", String.valueOf(colSpan));
            }
            int rowSpan = htmlAttributes.rowspan();
            if (rowSpan > 1) {
                printAttribute("rowspan", String.valueOf(rowSpan));
            }
            printAttribute("width", htmlAttributes.width());
        }
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _TH() {
        endTag("th");
        breakLine();
        return this;
    }

    public HtmlPrinter THEAD() {
        startTag("thead");
        breakLine();
        return this;
    }

    public HtmlPrinter THEAD(HtmlAttributes htmlAttributes) {
        openTag("thead");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _THEAD() {
        endTag("thead");
        breakLine();
        return this;
    }

    public HtmlPrinter TITLE() {
        startTag("title");
        return this;
    }

    public HtmlPrinter TITLE(HtmlAttributes htmlAttributes) {
        openTag("title");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        return this;
    }

    public HtmlPrinter _TITLE() {
        endTag("title");
        breakLine();
        return this;
    }

    public HtmlPrinter TR() {
        startTag("tr");
        breakLine();
        return this;
    }

    public HtmlPrinter TR(HtmlAttributes htmlAttributes) {
        openTag("tr");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _TR() {
        endTag("tr");
        breakLine();
        return this;
    }

    public HtmlPrinter UL() {
        startTag("ul");
        breakLine();
        return this;
    }

    public HtmlPrinter UL(HtmlAttributes htmlAttributes) {
        openTag("ul");
        printHTMLAttributes(htmlAttributes);
        closeTag();
        breakLine();
        return this;
    }

    public HtmlPrinter _UL() {
        endTag("ul");
        breakLine();
        return this;
    }


    private boolean printHTMLAttributes(HtmlAttributes htmlAttributes) {
        if (htmlAttributes == null) {
            return false;
        }
        printAttribute("class", htmlAttributes.classes());
        printAttribute("id", htmlAttributes.id());
        printAttribute("style", htmlAttributes.style());
        printAttribute("name", htmlAttributes.name());
        Map<String, String> attrMap = htmlAttributes.getAttrMap();
        if (attrMap != null) {
            for (Map.Entry<String, String> entry : attrMap.entrySet()) {
                pw.print(" ");
                escape(entry.getKey());
                pw.print("=\"");
                escape(entry.getValue());
                pw.print("\"");
            }
        }
        printAttribute("title", htmlAttributes.title());
        return true;
    }

    private void printAttribute(String name, String value) {
        if (value == null) {
            return;
        }
        pw.print(" ");
        pw.print(name);
        pw.print("=\"");
        escape(value);
        pw.print("\"");
    }

    private void printBooleanAttribute(String name, boolean value) {
        if (value) {
            printAttribute(name, name);
        }
    }

    private void startTag(String tag) {
        pw.print("<");
        pw.print(tag);
        pw.print(">");
    }

    private void openTag(String tag) {
        pw.print("<");
        pw.print(tag);
    }

    private void closeTag() {
        pw.print(">");
    }

    private void endTag(String tag) {
        pw.print("</");
        pw.print(tag);
        pw.print(">");
    }

    private void printPreformattedAttribute(String name, String value, boolean preformatted) {
        if (value == null) {
            return;
        }
        pw.print(" ");
        pw.print(name);
        pw.print("=\"");
        escape(value, preformatted);
        pw.print("\"");
    }

}
