/* UtilLib - Copyright (c) 2012-2017 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package net.mapeadores.util.logging;

import java.io.IOException;
import java.io.StringReader;
import java.util.AbstractList;
import java.util.Collections;
import java.util.List;
import java.util.RandomAccess;
import javax.xml.parsers.DocumentBuilder;
import net.mapeadores.util.exceptions.ShouldNotOccurException;
import net.mapeadores.util.html.HA;
import net.mapeadores.util.html.HtmlPrinter;
import net.mapeadores.util.instruction.InstructionErrorHandler;
import net.mapeadores.util.json.JSONWriter;
import net.mapeadores.util.localisation.LocalisationUtils;
import net.mapeadores.util.localisation.Message;
import net.mapeadores.util.xml.AppendableXMLWriter;
import net.mapeadores.util.xml.DOMUtils;
import net.mapeadores.util.xml.XMLUtils;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;


/**
 *
 * @author Vincent Calame
 */
public class LogUtils {

    public final static List<CategorizedMessageEntry> EMPTY_CATEGORIZEDMESSAGEENTRYLIST = Collections.emptyList();
    public final static MessageLog EMPTY_MESSAGELOG = new EmptyMessageLog();
    public final static List<MessageLog.LogGroup> EMPTY_LOGGROUPLIST = Collections.emptyList();
    public final static MultiMessageHandler NULL_MULTIMESSAGEHANDLER = new NullMultiMessageHandler();
    public final static LineMessageHandler NULL_LINEMESSAGEHANDLER = new NullLineMessageHandler();

    private LogUtils() {
    }

    public static String toString(String processURI, List<CategorizedMessageEntry> messageList) {
        StringBuilder buf = new StringBuilder();
        try {
            append(buf, processURI, messageList);
        } catch (IOException ioe) {
        }
        return buf.toString();
    }

    public static String toString(MessageLog messageLog) {
        StringBuilder buf = new StringBuilder();
        try {
            for (MessageLog.LogGroup group : messageLog.getLogGroupList()) {
                append(buf, group.getURI(), group.getEntryList());
            }
        } catch (IOException ioe) {
        }
        return buf.toString();
    }

    public static String toXmlString(MessageLog messageLog) {
        StringBuilder buf = new StringBuilder();
        try {
            AppendableXMLWriter xmlWriter = XMLUtils.toXMLWriter(buf);
            XMLUtils.appendXmlDeclaration(xmlWriter);
            MessageXMLPart messageXMLPart = new MessageXMLPart(xmlWriter);
            messageXMLPart.addLogs(messageLog);
        } catch (IOException ioe) {

        }
        return buf.toString();
    }

    public static String toXmlString(String uri, List<CategorizedMessageEntry> messageList) {
        StringBuilder buf = new StringBuilder();
        try {
            AppendableXMLWriter xmlWriter = XMLUtils.toXMLWriter(buf);
            XMLUtils.appendXmlDeclaration(xmlWriter);
            MessageXMLPart messageXMLPart = new MessageXMLPart(xmlWriter);
            messageXMLPart.addLogGroup(uri, messageList);
        } catch (IOException ioe) {

        }
        return buf.toString();
    }


    public static List<CategorizedMessageEntry> readCategorizedMessageList(String xmlString) {
        SimpleMessageHandler simpleHandler = new SimpleMessageHandler();
        DocumentBuilder docBuilder = DOMUtils.newDocumentBuilder();
        try {
            Document document = docBuilder.parse(new InputSource(new StringReader(xmlString)));
            MessageDOMReader reader = new MessageDOMReader(simpleHandler);
            return reader.readList(document.getDocumentElement(), "");
        } catch (SAXException saxe) {
            simpleHandler.addMessage("LOG XML FORMAT", LocalisationUtils.toMessage("_ error.exception.xml.sax", saxe.getMessage()));
            return simpleHandler.toList();
        } catch (IOException ioe) {
            throw new ShouldNotOccurException(ioe);
        }
    }

    public static MessageLog readMessageLog(String xmlString) {
        SimpleMessageHandler simpleHandler = new SimpleMessageHandler();
        MessageLogBuilder logBuilder = new MessageLogBuilder();
        DocumentBuilder docBuilder = DOMUtils.newDocumentBuilder();
        try {
            Document document = docBuilder.parse(new InputSource(new StringReader(xmlString)));
            MessageDOMReader reader = new MessageDOMReader(simpleHandler);
            reader.readProcessMessageLog(logBuilder, document.getDocumentElement(), "");
        } catch (SAXException saxe) {
            logBuilder.setCurrentURI("logfile");
            logBuilder.addMessage("LOG XML FORMAT", LocalisationUtils.toMessage("_ error.exception.xml.sax", saxe.getMessage()));
        } catch (IOException ioe) {
            throw new ShouldNotOccurException(ioe);
        }
        return logBuilder.toMessageLog();
    }

    public static void printHtml(HtmlPrinter hp, MessageLog messageLog) {
        hp
                .UL();
        for (MessageLog.LogGroup group : messageLog.getLogGroupList()) {
            List<CategorizedMessageEntry> messageList = group.getEntryList();
            if (!messageList.isEmpty()) {
                hp
                        .LI();
                hp
                        .P()
                        .CODE(HA.classes("cm-link"))
                        .escape(group.getURI())
                        ._CODE()
                        ._P();
                hp
                        .UL();
                for (CategorizedMessageEntry entry : messageList) {
                    Message message = entry.getMessage();
                    String category = entry.getCategory();
                    hp
                            .LI()
                            .P()
                            .CODE(HA.classes("cm-qualifier"))
                            .escape('[')
                            .escape(category)
                            .escape(']')
                            ._CODE()
                            .space()
                            .SMALL()
                            .localize(message)
                            ._SMALL()
                            ._P()
                            ._LI();
                }
                hp
                        ._UL();
                hp
                        ._LI();
            }
        }
        hp
                ._UL();
    }

    public static void writeJsonValue(JSONWriter jsonWriter, MessageLog messageLog) throws IOException {
        jsonWriter.array();
        for (MessageLog.LogGroup group : messageLog.getLogGroupList()) {
            List<CategorizedMessageEntry> messageList = group.getEntryList();
            if (!messageList.isEmpty()) {
                jsonWriter.object();
                jsonWriter.key("uri");
                jsonWriter.value(group.getURI());
                jsonWriter.key("messageArray");
                jsonWriter.array();
                for (CategorizedMessageEntry entry : messageList) {
                    Message message = entry.getMessage();
                    String category = entry.getCategory();
                    jsonWriter.object();
                    jsonWriter.key("category");
                    jsonWriter.value(category);
                    jsonWriter.key("key");
                    jsonWriter.value(message.getMessageKey());
                    Object[] values = message.getMessageValues();
                    if (values.length > 0) {
                        jsonWriter.key("valueArray");
                        jsonWriter.array();
                        for (Object obj : values) {
                            jsonWriter.value(obj.toString());
                        }
                        jsonWriter.endArray();
                    }
                    jsonWriter.endObject();
                }
                jsonWriter.endArray();
                jsonWriter.endObject();
            }
        }
        jsonWriter.endArray();
    }

    public static void append(Appendable buf, String processURI, List<CategorizedMessageEntry> messageList) throws IOException {
        if (messageList.isEmpty()) {
            return;
        }
        boolean withTabs = false;
        if (processURI != null) {
            buf.append("[[");
            buf.append(processURI);
            buf.append("]]");
            buf.append("\n");
            withTabs = true;
        }
        for (CategorizedMessageEntry entry : messageList) {
            Message message = entry.getMessage();
            String category = entry.getCategory();
            if (withTabs) {
                buf.append("      ");
            }
            buf.append("[");
            buf.append(category);
            buf.append("] ");
            buf.append(message.getMessageKey());
            Object[] values = message.getMessageValues();
            int length = values.length;
            if (length > 0) {
                buf.append(" = ");
                for (int i = 0; i < length; i++) {
                    if (i > 0) {
                        buf.append(" | ");
                    }
                    buf.append(values[i].toString());
                }
            }
            buf.append("\n");
        }
    }

    public static InstructionErrorHandler encapsulate(LineMessageHandler lineMessageHandler, String category, int lineNumber) {
        return new LineInstructionErrorHandler(lineMessageHandler, category, lineNumber);
    }

    public static List<MessageLog.LogGroup> wrap(MessageLog.LogGroup[] array) {
        return new MessageLogGroupList(array);
    }

    public static List<CategorizedMessageEntry> wrap(CategorizedMessageEntry[] array) {
        return new CategorizedMessageEntryList(array);
    }

    public static List<MessageByLine> wrap(MessageByLine[] array) {
        return new MessageByLineList(array);
    }


    private static class MessageLogGroupList extends AbstractList<MessageLog.LogGroup> implements RandomAccess {

        private final MessageLog.LogGroup[] array;

        private MessageLogGroupList(MessageLog.LogGroup[] array) {
            this.array = array;
        }

        @Override
        public int size() {
            return array.length;
        }

        @Override
        public MessageLog.LogGroup get(int index) {
            return array[index];
        }

    }


    private static class EmptyMessageLog implements MessageLog {

        private EmptyMessageLog() {
        }

        @Override
        public List<LogGroup> getLogGroupList() {
            return EMPTY_LOGGROUPLIST;
        }

    }


    private static class NullMultiMessageHandler implements MultiMessageHandler {

        private NullMultiMessageHandler() {
        }

        @Override
        public void addMessage(String category, Message message) {
        }

        @Override
        public void setCurrentURI(String uri) {
        }

    }


    private static class NullLineMessageHandler implements LineMessageHandler {

        private NullLineMessageHandler() {
        }

        @Override
        public void addMessage(int lineNumber, String category, Message message) {
        }


    }


    private static class LineInstructionErrorHandler implements InstructionErrorHandler {

        private final String category;
        private final LineMessageHandler lineMessageHandler;
        private final int lineNumber;

        private LineInstructionErrorHandler(LineMessageHandler lineMessageHandler, String category, int lineNumber) {
            this.lineMessageHandler = lineMessageHandler;
            this.category = category;
            this.lineNumber = lineNumber;
        }

        @Override
        public void invalidAsciiCharacterError(String part, int row, int col) {
            addMessage("_ error.wrong.character_notascii", part, row, col);
        }

        @Override
        public void invalidEndCharacterError(String part, int row, int col) {
            addMessage("_ error.wrong.character_end", part, row, col);
        }

        @Override
        public void invalidSeparatorCharacterError(String part, int row, int col) {
            addMessage("_ error.wrong.character_separator", part, row, col);
        }

        private void addMessage(String messageKey, String part, int row, int col) {
            lineMessageHandler.addMessage(lineNumber, category, LocalisationUtils.toMessage(messageKey, part, row, col));
        }

    }


    private static class CategorizedMessageEntryList extends AbstractList<CategorizedMessageEntry> implements RandomAccess {

        private final CategorizedMessageEntry[] array;

        private CategorizedMessageEntryList(CategorizedMessageEntry[] array) {
            this.array = array;
        }

        @Override
        public int size() {
            return array.length;
        }

        @Override
        public CategorizedMessageEntry get(int index) {
            return array[index];
        }

    }


    private static class MessageByLineList extends AbstractList<MessageByLine> implements RandomAccess {

        private final MessageByLine[] array;

        private MessageByLineList(MessageByLine[] array) {
            this.array = array;
        }

        @Override
        public int size() {
            return array.length;
        }

        @Override
        public MessageByLine get(int index) {
            return array[index];
        }

    }

}
